#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-11-03
# @Filename: spaxel.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-08 18:00:34


from __future__ import absolute_import, division, print_function

import inspect
import itertools
import warnings

import numpy as np

import marvin
import marvin.core.exceptions
import marvin.core.marvin_pickle
import marvin.tools.cube
import marvin.tools.maps
import marvin.tools.modelcube
import marvin.utils.general.general
from marvin.core.exceptions import MarvinBreadCrumb, MarvinError, MarvinUserWarning
from marvin.utils.datamodel.dap import datamodel as dap_datamodel
from marvin.utils.datamodel.drp import datamodel as drp_datamodel
from marvin.utils.general.structs import FuzzyDict


breadcrumb = MarvinBreadCrumb()


class DataModel(object):
    """A single object that holds the DRP and DAP datamodel."""

    def __init__(self, release):

        self.drp = drp_datamodel[release]
        self.dap = dap_datamodel[release]


class Spaxel(object):
    """A base class that contains information about a spaxel.

    This class represents an spaxel with information from the reduced DRP
    spectrum, the DAP maps properties, and the model spectrum from the DAP
    logcube. A `.SpaxelBase` can be initialised with all or only part of that
    information, and either from a file, a database, or remotely via the
    Marvin API.

    The `~marvin.tools.cube.Cube`, `~marvin.tools.maps.Maps` , and
    `~marvin.tools.modelcube.ModelCube` quantities for the spaxel are available
    in ``cube_quantities``, ``maps_quantities``, and ``modelcube_quantities``,
    respectively. For convenience, the quantities can also be accessed directly
    from the `.SpaxelBase` itself (e.g., ``spaxel.emline_gflux_ha_6465``).

    Parameters:
        x,y (int):
            The `x` and `y` coordinates of the spaxel in the cube (0-indexed).
        cube (`~marvin.tools.cube.Cube` object or path or bool):
            If ``cube`` is a `~marvin.tools.cube.Cube` object, that
            cube will be used for the `.SpaxelBase` instantiation. This mode
            is mostly intended for `~marvin.utils.general.general.getSpaxel`
            as it significantly improves loading time. Otherwise, ``cube`` can
            be ``True`` (default), in which case a cube will be instantiated
            using the input ``filename``, ``mangaid``, or ``plateifu``. If
            ``cube=False``, no cube will be used and the cube associated
            quantities will not be available. ``cube`` can also be the
            path to the DRP cube to use.
        maps (`~marvin.tools.maps.Maps` object or path or bool)
            As ``cube`` but for the DAP measurements corresponding to the
            spaxel in the `.Maps`.
        modelcube (`marvin.tools.modelcube.ModelCube` object or path or bool)
            As ``maps`` but for the DAP measurements corresponding to the
            spaxel in the `.ModelCube`.
        lazy (bool):
            If ``False``, the spaxel data is loaded on instantiation.
            Otherwise, only the metadata is created. The associated quantities
            can be then loaded by calling `.SpaxelBase.load()`.
        kwargs (dict):
            Arguments to be passed to `.Cube`, `.Maps`, and `.ModelCube`
            when (and if) they are initialised.

    Attributes:
        cube_quantities (`~marvin.utils.general.structs.FuzzyDict`):
            A querable dictionary with the `.Spectrum` quantities
            derived from `.Cube` and matching ``x, y``.
        datamodel (object):
            An object containing the DRP and DAP datamodels.
        maps_quantities (`~marvin.utils.general.structs.FuzzyDict`):
            A querable dictionary with the `.AnalysisProperty` quantities
            derived from `.Maps` and matching ``x, y``.
        model_quantities (`~marvin.utils.general.structs.FuzzyDict`):
            A querable dictionary with the `.Spectrum` quantities
            derived from `.ModelCube` and matching ``x, y``.
        ra,dec (float):
            Right ascension and declination of the spaxel. Not available until
            the spaxel has been `loaded <.SpaxelBase.load>`.

    """

    def __init__(self, x, y, cube=True, maps=True, modelcube=True, lazy=False, **kwargs):

        if not cube and not maps and not modelcube:
            raise MarvinError('no inputs defined.')

        self.cube_quantities = FuzzyDict({})
        self.maps_quantities = FuzzyDict({})
        self.modelcube_quantities = FuzzyDict({})

        self._cube = cube
        self._maps = maps
        self._modelcube = modelcube

        for attr in ['mangaid', 'plateifu', 'release', 'bintype', 'template']:

            value = kwargs.pop(attr, None) or \
                getattr(cube, attr, None) or \
                getattr(maps, attr, None) or \
                getattr(modelcube, attr, None)

            setattr(self, attr, value)

        self._kwargs = kwargs
        self._parent_shape = None

        # drop breadcrumb
        breadcrumb.drop(message='Initializing MarvinSpaxel {0}'.format(self.__class__),
                        category=self.__class__)

        self.x = int(x)
        self.y = int(y)

        self.loaded = False
        self.datamodel = None

        if lazy is False:
            self.load()

        # Load VACs
        from marvin.contrib.vacs.base import VACMixIn
        self.vacs = VACMixIn.get_vacs(self)

    def __dir__(self):

        class_members = list(list(zip(*inspect.getmembers(self.__class__)))[0])
        instance_attr = list(self.__dict__.keys())

        items = self.cube_quantities.__dir__()
        items += self.maps_quantities.__dir__()
        items += self.modelcube_quantities.__dir__()
        items += class_members + instance_attr

        return sorted(items)

    def __getattr__(self, value):

        _getattr = super(Spaxel, self).__getattribute__

        for tool_quantity_dict in ['cube_quantities', 'maps_quantities', 'modelcube_quantities']:
            if value in _getattr(tool_quantity_dict):
                return _getattr(tool_quantity_dict)[value]

        return super(Spaxel, self).__getattribute__(value)

    def __repr__(self):
        """Spaxel representation."""

        if not self.loaded:
            return '<Marvin Spaxel (x={0.x:d}, y={0.y:d}, loaded=False)'.format(self)

        # Gets the coordinates relative to the centre of the cube/maps.
        y_mid, x_mid = np.array(self._parent_shape) / 2.
        x_centre = int(self.x - x_mid)
        y_centre = int(self.y - y_mid)

        # Determine what tools are loaded.
        tools = np.array(['cube', 'maps', 'modelcube'])
        load_idx = np.where([self._cube, self._maps, self._modelcube])[0]
        flags = '/'.join(tools[load_idx])

        return ('<Marvin Spaxel (plateifu={0.plateifu}, x={0.x:d}, y={0.y:d}; '
                'x_cen={1:d}, y_cen={2:d}, loaded={3})>'.format(self, x_centre, y_centre, flags))

    def _check_versions(self, attr):
        """Checks that all input object have the same versions.

        Runs sanity checks to make sure that ``attr`` has the same value
        in the input `.Cube`, `.Maps`, and `.ModelCube`. Returns the value
        for the attribute or ``None`` if the attribute does not exist.

        """

        out_value = None

        inputs = []
        for obj in [self._cube, self._maps, self._modelcube]:
            if obj is not None and not isinstance(obj, bool):
                inputs.append(obj)

        if len(inputs) == 1:
            return getattr(inputs[0], attr, None)

        for obj_a, obj_b in itertools.combinations(inputs, 2):
            if hasattr(obj_a, attr) and hasattr(obj_b, attr):
                assert getattr(obj_a, attr) == getattr(obj_b, attr), \
                    'inconsistent {!r} between {!r} and {!r}'.format(attr, obj_a, obj_b)
            out_value = getattr(obj_a, attr, None) or getattr(obj_b, attr, None)

        return out_value

    def _set_radec(self):
        """Calculates ra and dec for this spaxel."""

        self.ra = None
        self.dec = None

        for obj in [self._cube, self._maps, self._modelcube]:
            if hasattr(obj, 'wcs'):
                if obj.wcs.naxis == 2:
                    self.ra, self.dec = obj.wcs.wcs_pix2world([[self.x, self.y]], 0)[0]
                elif obj.wcs.naxis == 3:
                    self.ra, self.dec, __ = obj.wcs.wcs_pix2world([[self.x, self.y, 0]], 0)[0]

    def save(self, path, overwrite=False):
        """Pickles the spaxel to a file.

        Parameters:
            path (str):
                The path of the file to which the `.Spaxel` will be saved.
                Unlike for other Marvin Tools that derive from
                `~marvin.tools.core.MarvinToolsClass`, ``path`` is
                mandatory for `.Spaxel.save` as there is no default path for a
                given spaxel.
            overwrite (bool):
                If True, and the ``path`` already exists, overwrites it.
                Otherwise it will fail.

        Returns:
            path (str):
                The realpath to which the file has been saved.

        """

        return marvin.core.marvin_pickle.save(self, path=path, overwrite=overwrite)

    @classmethod
    def restore(cls, path, delete=False):
        """Restores a Spaxel object from a pickled file.

        If ``delete=True``, the pickled file will be removed after it has been
        unplickled. Note that, for objects with ``data_origin='file'``, the
        original file must exists and be in the same path as when the object
        was first created.

        """

        return marvin.core.marvin_pickle.restore(path, delete=delete)

    def load(self, force=None):
        """Loads the spaxel data.

        Loads the spaxel data for cubes/maps/modelcube. By default attempts
        to load whatever is specified when spaxels are instantianted from other
        Marvin Tools. Can manually force load a data type with the force
        keyword.

        Parameters:
        -----------
        force : {cube|maps|modelcube}
            Datatype to force load.

        """

        if self.loaded and force is None:
            warnings.warn('already loaded', MarvinUserWarning)
            return

        assert force in [None, 'cube', 'maps', 'modelcube'], \
            'force can only be cube, maps, or modelcube'

        for tool in ['cube', 'maps', 'modelcube']:
            self._load_tool(tool, force=(force is not None and force == tool))

        self._set_radec()
        self.loaded = True

        for attr in ['mangaid', 'plateifu', 'release', 'bintype', 'template']:
            setattr(self, attr, self._check_versions(attr))

        self.datamodel = DataModel(self.release)

    def _load_tool(self, tool, force=False):
        """Loads the tool and the associated quantities."""

        if tool == 'cube':
            class_name = marvin.tools.cube.Cube
            method = self.getCube
            quantities_dict = 'cube_quantities'
        elif tool == 'maps':
            class_name = marvin.tools.maps.Maps
            method = self.getMaps
            quantities_dict = 'maps_quantities'
        elif tool == 'modelcube':
            class_name = marvin.tools.modelcube.ModelCube
            method = self.getModelCube
            quantities_dict = 'modelcube_quantities'

        attr_value = getattr(self, '_' + tool)
        if (attr_value is False or attr_value is None) and force is False:
            setattr(self, '_' + tool, None)
            return

        if not isinstance(attr_value, class_name):
            if tool == 'modelcube' and self.release == 'MPL-4':
                warnings.warn('ModelCube cannot be instantiated for MPL-4.', MarvinUserWarning)
                self._modelcube = None
                return
            else:
                setattr(self, '_' + tool, method())
        else:
            if force is True:
                warnings.warn('{0} is already loaded'.format(tool), MarvinUserWarning)

        self._parent_shape = getattr(getattr(self, '_' + tool), '_shape')

        setattr(self, quantities_dict,
                getattr(getattr(self, '_' + tool), '_get_spaxel_quantities')(self.x, self.y,
                                                                             spaxel=self))

    def getCube(self):
        """Returns the associated `~marvin.tools.cube.Cube`"""

        if isinstance(self._cube, marvin.tools.cube.Cube):
            return self._cube

        cube_input = (self._cube if self._cube is not True else None) \
            or self.plateifu or self.mangaid

        return marvin.tools.cube.Cube(cube_input, release=self.release, **self._kwargs)

    def getMaps(self):
        """Returns the associated `~marvin.tools.maps.Maps`"""

        if isinstance(self._maps, marvin.tools.maps.Maps):
            return self._maps

        maps_input = (self._maps if self._maps is not True else None) \
            or self.plateifu or self.mangaid

        return marvin.tools.maps.Maps(maps_input, bintype=self.bintype,
                                      template=self.template, release=self.release,
                                      **self._kwargs)

    def getModelCube(self):
        """Returns the associated `~marvin.tools.modelcube.ModelCube`"""

        if isinstance(self._modelcube, marvin.tools.modelcube.ModelCube):
            return self._modelcube

        modelcube_input = (self._modelcube if self._modelcube is not True else None) \
            or self.plateifu or self.mangaid

        return marvin.tools.modelcube.ModelCube(modelcube_input, bintype=self.bintype,
                                                template=self.template, release=self.release,
                                                **self._kwargs)

    @property
    def quality_flags(self):
        """Bundle Cube DRP3QUAL and Maps DAPQUAL flags."""

        drp3qual = self.datamodel.drp.bitmasks['MANGA_DRP3QUAL']
        cube = self.getCube()
        drp3qual.mask = int(cube.header['DRP3QUAL'])
        qual_flags = [drp3qual]

        if self.release != 'MPL-4':
            qual_flags.append(self.datamodel.dap.bitmasks['MANGA_DAPQUAL'])

        return qual_flags



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Brian Cherinka, José Sánchez-Gallego, and Brett Andrews
# @Date: 2017-10-25
# @Filename: base.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)
#
# @Last modified by:   Brian Cherinka
# @Last modified time: 2018-07-25 18:32:35

from __future__ import absolute_import, division, print_function

import copy as copy_mod
import os

import astropy.table as table
from astropy import units as u

from marvin.core.exceptions import MarvinError
from marvin.utils.datamodel import DataModelList
from marvin.utils.general.structs import FuzzyList


class DRPCubeDataModel(object):
    """A class representing a DRP Cube datamodel.

    Parameters
    ----------
    release : str
        The DRP release this datamodel describes.
    datacubes : list
        A list of `.DataCube` instances that describe the datacubes in this
        datamodel.
    spectra : list
        A list of `.Spectrum` instances that describe the datacubes in this
        datamodel.
    aliases : list
        A list of aliases for this datamodel.
    bitmask : dict
        A dictionary of `~marvin.utils.general.maskbit.Maskbit` objects.
    qual_flag : str
        The name of the quality bitmask flag. Must not include the ``MANGA_``
        prefix.

    """

    def __init__(self, release, datacubes=[], spectra=[], aliases=[], bitmasks=None,
                 qual_flag='DRP3QUAL'):

        self.release = release
        self.aliases = aliases

        self.datacubes = DataCubeList(datacubes, parent=self)
        self.spectra = SpectrumList(spectra, parent=self)

        self.bitmasks = bitmasks if bitmasks is not None else {}
        self.qual_flag = qual_flag

    def __repr__(self):
        return ('<DRPCubeDataModel release={0!r}, n_datacubes={1}, n_spectra={2}>'
                .format(self.release, len(self.datacubes), len(self.spectra)))

    def copy(self):
        """Returns a copy of the datamodel."""

        return copy_mod.deepcopy(self)

    def __eq__(self, value):
        """Uses fuzzywuzzy to return the closest property match."""

        datacube_names = [datacube.name for datacube in self.datacubes]
        spectrum_names = [spectrum.name for spectrum in self.spectra]

        if value in datacube_names:
            return self.datacubes[datacube_names.index(value)]
        elif value in spectrum_names:
            return self.spectra[spectrum_names.index(value)]

        try:
            datacube_best_match = self.datacubes[value]
        except ValueError:
            datacube_best_match = None

        try:
            spectrum_best_match = self.spectra[value]
        except ValueError:
            spectrum_best_match = None

        if ((datacube_best_match is None and spectrum_best_match is None) or
                (datacube_best_match is not None and spectrum_best_match is not None)):
            raise ValueError('too ambiguous input {!r}'.format(value))
        elif datacube_best_match is not None:
            return datacube_best_match
        elif spectrum_best_match is not None:
            return spectrum_best_match

    def __contains__(self, value):

        try:
            match = self.__eq__(value)
            if match is None:
                return False
            else:
                return True
        except ValueError:
            return False

    def __getitem__(self, value):
        return self == value

    def to_rss(self):
        """Returns a copy with `.RSSDatamodel` objects instead of datacubes."""

        if isinstance(self, DRPRSSDataModel):
            raise ValueError('this is already a DRPRSSDataModel')

        copy_of_self = self.copy()
        delattr(copy_of_self, 'datacubes')

        # Chances the type and converts datacubes to RSSDatamodel
        copy_of_self.__class__ = DRPRSSDataModel
        copy_of_self.rss = self.datacubes.to_rss(copy_of_self)

        # Resets the parent of the copied spectra
        for spectrum in copy_of_self.spectra:
            spectrum.parent = copy_of_self

        return copy_of_self


class DRPRSSDataModel(DRPCubeDataModel):
    """A class representing a DRP RSS listdatamodel."""

    def __init__(self, release, rss=[], spectra=[], aliases=[], bitmasks=None):

        self.release = release
        self.aliases = aliases

        self.rss = RSSList(rss, parent=self)
        self.spectra = SpectrumList(spectra, parent=self)

        self.bitmasks = bitmasks if bitmasks is not None else {}

    def __repr__(self):
        return ('<DRPRSSDataModel release={0!r}, n_rss={1}, n_spectra={2}>'
                .format(self.release, len(self.rss), len(self.spectra)))


class DRPCubeDataModelList(DataModelList):
    """A dictionary of DRP Cube datamodels."""

    base = {'DRPCubeDataModel': DRPCubeDataModel}

    def copy(self):

        copy_of_self = super(DRPCubeDataModelList, self).copy()
        copy_of_self.__class__ = DRPRSSDataModelList

        return copy_of_self


class DRPRSSDataModelList(DataModelList):
    """A dictionary of DRP RSS datamodels."""

    base = {'DRPRSSDataModel': DRPRSSDataModel}


class DataCubeList(FuzzyList):
    """Creates a list containing models and their representation."""

    def __init__(self, the_list, parent=None):

        self.parent = parent

        super(DataCubeList, self).__init__([])

        for item in the_list:
            self.append(item, copy=True)

    def copy(self):
        """Returns a copy of the datamodel."""

        return copy_mod.deepcopy(self)

    def mapper(self, value):
        """Helper method for the fuzzy list to match on the datacube name."""

        return value.name

    def append(self, value, copy=True):
        """Appends with copy."""

        append_obj = value if copy is False else copy_mod.deepcopy(value)
        append_obj.parent = self.parent

        if isinstance(append_obj, DataCube):
            super(DataCubeList, self).append(append_obj)
        else:
            raise ValueError('invalid datacube of type {!r}'.format(type(append_obj)))

    def to_rss(self, new_parent):
        """Returns a copy of this list as an `.RSSList` object."""

        if isinstance(self, RSSList):
            raise ValueError('this is already an RSSList')

        # Copies selef and resets the type to RSSList
        copy_of_self = self.copy()
        copy_of_self.__class__ = RSSList
        copy_of_self.parent = new_parent

        # Replaces each datacube in itself with a RSSDatamodel
        for label in self.list_names():
            copy_of_self.remove(copy_of_self[label])
            copy_of_self.append(self[label].to_rss(new_parent))

        return copy_of_self

    def list_names(self):
        """Returns a list with the names of the datacubes in this list."""

        return [item.name for item in self]

    def to_table(self, pprint=False, description=False, max_width=1000):
        """Returns an astropy table with all the datacubes in this datamodel.

        Parameters:
            pprint (bool):
                Whether the table should be printed to screen using astropy's
                table pretty print.
            description (bool):
                If ``True``, an extra column with the description of the
                datacube will be added.
            max_width (int or None):
                A keyword to pass to ``astropy.table.Table.pprint()`` with the
                maximum width of the table, in characters.

        Returns:
            result (``astropy.table.Table``):
                If ``pprint=False``, returns an astropy table containing
                the name of the datacube, whether it has ``ivar`` or
                ``mask``, the units, and a description (if
                ``description=True``)..

        """

        datacube_table = table.Table(
            None, names=['name', 'ivar', 'mask', 'unit', 'description',
                         'db_table', 'db_column', 'fits_extension'],
            dtype=['S20', bool, bool, 'S20', 'S500', 'S20', 'S20', 'S20'])

        if self.parent:
            datacube_table.meta['release'] = self.parent.release

        for datacube in self:
            unit = datacube.unit.to_string()

            datacube_table.add_row((datacube.name,
                                    datacube.has_ivar(),
                                    datacube.has_mask(),
                                    unit,
                                    datacube.description,
                                    datacube.db_table,
                                    datacube.db_column(),
                                    datacube.fits_extension()))

        if not description:
            datacube_table.remove_column('description')

        if pprint:
            datacube_table.pprint(max_width=max_width, max_lines=1e6)
            return

        return datacube_table

    def write_csv(self, filename=None, path=None, overwrite=None, **kwargs):
        ''' Write the datamodel to a CSV '''

        release = self.parent.release.lower().replace('-', '')

        if not filename:
            filename = 'drpcubes_dm_{0}.csv'.format(release)

        if not path:
            path = os.path.join(os.getenv("MARVIN_DIR"), 'docs', 'sphinx', '_static')

        fullpath = os.path.join(path, filename)
        table = self.to_table(**kwargs)
        table.write(fullpath, format='csv', overwrite=overwrite)


class RSSList(DataCubeList):
    """Creates a list containing RSSDatamodel and their representation."""

    def append(self, value, copy=True):
        """Appends with copy."""

        append_obj = value if copy is False else copy_mod.deepcopy(value)
        append_obj.parent = self.parent

        if isinstance(append_obj, RSS):
            super(RSSList, self).append(append_obj)
        else:
            raise ValueError('invalid RSS of type {!r}'.format(type(append_obj)))

    def write_csv(self, filename=None, path=None, overwrite=None, **kwargs):
        """Write the datamodel to a CSV"""

        release = self.parent.release.lower().replace('-', '')

        if not filename:
            filename = 'drprss_dm_{0}.csv'.format(release)

        if not path:
            path = os.path.join(os.getenv('MARVIN_DIR'), 'docs', 'sphinx', '_static')

        fullpath = os.path.join(path, filename)
        table = self.to_table(**kwargs)
        table.write(fullpath, format='csv', overwrite=overwrite)


class DataCube(object):
    """Represents a extension in the DRP logcube file.

    Parameters:
        name (str):
            The datacube name. This is the internal name that Marvin will use
            for this datacube. It is different from the ``extension_name``
            parameter, which must be identical to the extension name of the
            datacube in the logcube file.
        extension_name (str):
            The FITS extension containing this datacube.
        extension_wave (str):
            The FITS extension containing the wavelength for this datacube.
        extension_ivar (str or None):
            The extension that contains the inverse variance associated with
            this datacube, if any.
        extension_mask (str or None):
            The extension that contains the mask associated with this
            datacube, if any.
        db_table (str):
            The DB table in which the datacube is stored. Defaults to
            ``spaxel``.
        db_column (str):
            An alternate DB column in which the datacube is stored. If none, defaults to
            the FITS extension name.
        unit (astropy unit or None):
            The unit for this datacube.
        scale (float):
            The scaling factor for the values of the datacube.
        formats (dict):
            A dictionary with formats that can be used to represent the
            datacube. Default ones are ``latex`` and ``string``.
        pixmask_flag : str
            The name of the pixmask flag. Should be the full name, including the
            ``MANGA_`` part.
        description (str):
            A description for the datacube.

    """

    def __init__(self, name, extension_name, extension_wave=None,
                 extension_ivar=None, extension_mask=None, db_table='spaxel',
                 db_column=None, unit=u.dimensionless_unscaled, scale=1, formats={},
                 pixmask_flag='MANGA_DRP3PIXMASK', description=''):

        self.name = name

        self._extension_name = extension_name
        self._extension_wave = extension_wave
        self._extension_ivar = extension_ivar
        self._extension_mask = extension_mask

        self.pixmask_flag = pixmask_flag

        self.db_table = db_table
        self._db_column = db_column

        self._parent = None

        self.formats = formats

        self.description = description

        self.unit = u.CompositeUnit(scale, unit.bases, unit.powers)

    def copy(self):
        return copy_mod.deepcopy(self)

    def to_rss(self, new_parent):
        """Creates a copy of this datacube as a `.RSS` object."""

        if isinstance(self, RSS):
            raise ValueError('this is already a RSS datamodel object.')

        assert isinstance(new_parent, DRPRSSDataModel)
        copy_of_self = self.copy()
        copy_of_self.__class__ = RSS
        copy_of_self.parent = new_parent
        copy_of_self.db_table = 'rssfiber'

        return copy_of_self

    @property
    def parent(self):
        """Retrieves the parent."""

        return self._parent

    @parent.setter
    def parent(self, value):
        """Sets the parent."""

        assert isinstance(value, DRPCubeDataModel), 'parent must be a DRPCubeDataModel'

        self._parent = value

    def full(self):
        """Returns the name string."""

        return self._extension_name.lower()

    def has_ivar(self):
        """Returns True is the datacube has an ivar extension."""

        return self._extension_ivar is not None

    def has_mask(self):
        """Returns True is the datacube has an mask extension."""

        return self._extension_mask is not None

    def fits_extension(self, ext=None):
        """Returns the FITS extension name."""

        assert ext is None or ext in ['ivar', 'mask'], 'invalid extension'

        if ext is None:
            return self._extension_name.upper()

        elif ext == 'ivar':
            if not self.has_ivar():
                raise MarvinError('no ivar extension for datacube {0!r}'.format(self.full()))
            return self._extension_ivar.upper()

        elif ext == 'mask':
            if not self.has_mask():
                raise MarvinError('no mask extension for datacube {0!r}'.format(self.full()))
            return self._extension_mask

    def db_column(self, ext=None):
        """Returns the name of the DB column containing this datacube.

        If ``db_column`` is passed in as input to the ``DataCube`` datamodel, returns
        the given name.  Otherwise returns the name of the FITS extension.
        """

        if self._db_column:
            return self._db_column
        return self.fits_extension(ext=ext).lower()

    def __repr__(self):

        return '<DataCube {!r}, release={!r}, unit={!r}>'.format(
            self.name, self.parent.release if self.parent else None, self.unit.to_string())

    def __str__(self):

        return self.full()

    def to_string(self, mode='string'):
        """Return a string representation of the datacube."""

        if mode == 'latex':

            if mode in self.formats:
                latex = self.formats[mode]
            else:
                latex = self.to_string()

            return latex

        else:

            if mode in self.formats:
                string = self.formats[mode]
            else:
                string = self.name

            return string


class RSS(DataCube):

    def __repr__(self):

        return '<RSS {!r}, release={!r}, unit={!r}>'.format(
            self.name, self.parent.release if self.parent else None,
            self.unit.to_string())

    @property
    def parent(self):
        """Retrieves the parent."""

        return self._parent

    @parent.setter
    def parent(self, value):
        """Sets the parent."""

        assert isinstance(value, DRPRSSDataModel), 'parent must be a DRPRSSDataModel'
        self._parent = value


class SpectrumList(FuzzyList):
    """Creates a list containing spectra and their representation."""

    def __init__(self, the_list, parent=None):

        self.parent = parent

        super(SpectrumList, self).__init__([])

        for item in the_list:
            self.append(item, copy=True)

    def mapper(self, value):
        """Helper method for the fuzzy list to match on the spectrum name."""

        return value.name

    def append(self, value, copy=True):
        """Appends with copy."""

        append_obj = value if copy is False else copy_mod.deepcopy(value)
        append_obj.parent = self.parent

        if isinstance(append_obj, Spectrum):
            super(SpectrumList, self).append(append_obj)
        else:
            raise ValueError('invalid spectrum of type {!r}'.format(type(append_obj)))

    def list_names(self):
        """Returns a list with the names of the spectra in this list."""

        return [item.name for item in self]

    def to_table(self, pprint=False, description=False, max_width=1000):
        """Returns an astropy table with all the spectra in this datamodel.

        Parameters:
            pprint (bool):
                Whether the table should be printed to screen using astropy's
                table pretty print.
            description (bool):
                If ``True``, an extra column with the description of the
                spectrum will be added.
            max_width (int or None):
                A keyword to pass to ``astropy.table.Table.pprint()`` with the
                maximum width of the table, in characters.

        Returns:
            result (``astropy.table.Table``):
                If ``pprint=False``, returns an astropy table containing
                the name of the spectrum, whether it has ``ivar`` or
                ``mask``, the units, and a description (if
                ``description=True``)..

        """

        spectrum_table = table.Table(
            None, names=['name', 'std', 'unit', 'description',
                         'db_table', 'db_column', 'fits_extension'],
            dtype=['S20', bool, 'S20', 'S500', 'S20', 'S20', 'S20'])

        if self.parent:
            spectrum_table.meta['release'] = self.parent.release

        for spectrum in self:
            unit = spectrum.unit.to_string()

            spectrum_table.add_row((spectrum.name,
                                    spectrum.has_std(),
                                    unit,
                                    spectrum.description,
                                    spectrum.db_table,
                                    spectrum.db_column(),
                                    spectrum.fits_extension()))

        if not description:
            spectrum_table.remove_column('description')

        if pprint:
            spectrum_table.pprint(max_width=max_width, max_lines=1e6)
            return

        return spectrum_table

    def write_csv(self, filename=None, path=None, overwrite=None, **kwargs):
        ''' Write the datamodel to a CSV '''

        release = self.parent.release.lower().replace('-', '')

        if not filename:
            if isinstance(self.parent, DRPRSSDataModel):
                filename = 'drp_rss_spectra_dm_{0}.csv'.format(release)
            elif isinstance(self.parent, DRPCubeDataModel):
                filename = 'drp_cube_spectra_dm_{0}.csv'.format(release)
            else:
                raise ValueError('invalid parent of type {!r}'.format(type(self.parent)))

        if not path:
            path = os.path.join(os.getenv("MARVIN_DIR"), 'docs', 'sphinx', '_static')

        fullpath = os.path.join(path, filename)
        table = self.to_table(**kwargs)
        table.write(fullpath, format='csv', overwrite=overwrite)


class Spectrum(object):
    """Represents a extension in the DRP logcube file.

    Parameters:
        name (str):
            The spectrum name. This is the internal name that Marvin will use
            for this spectrum. It is different from the ``extension_name``
            parameter, which must be identical to the extension name of the
            spectrum in the logcube file.
        extension_name (str):
            The FITS extension containing this spectrum.
        extension_wave (str):
            The FITS extension containing the wavelength for this spectrum.
        extension_std (str):
            The FITS extension containing the standard deviation for this
            spectrum.
        extension_mask (str):
            The FITS extension containing the mask for this spectrum.
        db_table (str):
            The DB table in which the spectrum is stored. Defaults to
            ``cube``.
        unit (astropy unit or None):
            The unit for this spectrum.
        scale (float):
            The scaling factor for the values of the spectrum.
        formats (dict):
            A dictionary with formats that can be used to represent the
            spectrum. Default ones are ``latex`` and ``string``.
        pixmask_flag : str
            The name of the pixmask flag. Should be the full name, including the
            ``MANGA_`` part.
        description (str):
            A description for the spectrum.

    """

    def __init__(self, name, extension_name, extension_wave=None, extension_std=None,
                 extension_mask=None, db_table='cube', unit=u.dimensionless_unscaled,
                 scale=1, formats={}, pixmask_flag=None, description=''):

        self.name = name

        self._extension_name = extension_name
        self._extension_wave = extension_wave
        self._extension_std = extension_std
        self._extension_mask = extension_mask

        self.pixmask_flag = pixmask_flag

        self.db_table = db_table

        self.formats = formats

        self.description = description

        self._parent = None

        self.unit = u.CompositeUnit(scale, unit.bases, unit.powers)

    @property
    def parent(self):
        """Retrieves the parent."""

        return self._parent

    @parent.setter
    def parent(self, value):
        """Sets the parent."""

        assert isinstance(value, DRPCubeDataModel), 'parent must be a DRPCubeDataModel'

        self._parent = value

    def full(self):
        """Returns the name string."""

        return self._extension_name.lower()

    def has_std(self):
        """Returns True is the datacube has an std extension."""

        return self._extension_std is not None

    def has_mask(self):
        """Returns True is the datacube has an mask extension."""

        return self._extension_mask is not None

    def fits_extension(self, ext=None):
        """Returns the FITS extension name."""

        assert ext is None or ext in ['std', 'mask'], 'invalid extension'

        if ext is None:
            return self._extension_name.upper()

        elif ext == 'std':
            if not self.has_std():
                raise MarvinError('no std extension for spectrum {0!r}'.format(self.full()))
            return self._extension_std.upper()

        elif ext == 'mask':
            if not self.has_mask():
                raise MarvinError('no mask extension for spectrum {0!r}'.format(self.full()))
            return self._extension_mask

    def db_column(self, ext=None):
        """Returns the name of the DB column containing this datacube."""

        return self.fits_extension(ext=ext).lower()

    def __repr__(self):

        return '<Spectrum {!r}, release={!r}, unit={!r}>'.format(
            self.name, self.parent.release if self.parent else None, self.unit.to_string())

    def __str__(self):

        return self.full()

    def to_string(self, mode='string'):
        """Return a string representation of the spectrum."""

        if mode == 'latex':

            if mode in self.formats:
                latex = self.formats[mode]
            else:
                latex = self.to_string()

            return latex

        else:

            if mode in self.formats:
                string = self.formats[mode]
            else:
                string = self.name

            return string



#!/usr/bin/env python
# encoding: utf-8
#
# maskbit.py
#
# @Author: Brett Andrews <andrews>
# @Date:   2017-10-06 10:10:00
# @Last modified by: José Sánchez-Gallego (gallegoj@uw.edu)
# @Last modified time: 2018-11-26 11:51:50

from __future__ import absolute_import, division, print_function

import os

import numpy as np
import pandas as pd

import marvin
from marvin.utils.yanny import yanny


# Stores the maskbits yanny file structure so that we don't need to open it more than once.
_maskbits_from_yanny = None


def _read_maskbit_schemas():
    """Read all available SDSS maskbit schemas from yanny file.

    Returns:
        Record Array: all bits for all schemas.
    """

    global _maskbits_from_yanny

    if _maskbits_from_yanny is None:
        path_maskbits = os.path.join(os.path.dirname(marvin.__file__), 'data', 'sdssMaskbits.par')
        _maskbits_from_yanny = yanny(path_maskbits, np=True)

    return _maskbits_from_yanny['MASKBITS']


def get_available_maskbits():
    """Get names of available maskbit schemas from yanny file.

    Returns:
        list: Names of available maskbits.
    """
    maskbits = _read_maskbit_schemas()
    return sorted(set([it[0] for it in maskbits]))


def get_manga_target(flag_id, bitmasks, header):
    """Get MANGA_TARGET[``flag_id``] flag.

    Parameters:
        flag_id (str):
            Flag ID number (e.g., "1" for MANGA_TARGET1).
        bitmasks (dict):
            `Maskbit` objects.
        header (`astropy.io.fits.header.Header`):
            File header.

    Returns:
        `Maskbit`
    """
    flag_id = str(int(flag_id))
    manga_target = bitmasks['MANGA_TARGET{}'.format(flag_id)]

    try:
        manga_target.mask = int(header['MNGTRG{}'.format(flag_id)])
    except KeyError:
        manga_target.mask = int(header['MNGTARG{}'.format(flag_id)])

    return manga_target


class Maskbit(object):
    """A class representing a maskbit.

    Parameters:
        schema (DataFrame):
            Maskbit schema.
        name (str):
            Name of maskbit.
        description (str):
            Description of maskbit.
    """

    def __init__(self, name, schema=None, description=None):

        self.name = name
        self.schema = schema if schema is not None else self._load_schema(name)
        self.description = description if description is not None else None
        self.mask = None

    def __repr__(self):
        if (isinstance(self.mask, int) or self.mask is None):
            labels = self.labels
        else:
            labels = 'shape={}'.format(self.mask.shape)
        return '<Maskbit {0!r} {1}>'.format(self.name, labels)

    def _load_schema(self, flag_name):
        """Load SDSS Maskbit schema from yanny file.

        Parameters:
            flag_name (str):
                Name of flag.

        Returns:
            DataFrame: Schema of flag.
        """
        maskbits = _read_maskbit_schemas()
        flag = maskbits[maskbits['flag'] == flag_name]

        return pd.DataFrame(flag[['bit', 'label', 'description']])

    @property
    def bits(self):
        return self.values_to_bits() if self.mask is not None else None

    @property
    def labels(self):
        return self.values_to_labels() if self.mask is not None else None

    def values_to_bits(self, values=None):
        """Convert mask values to a list of bits set.

        Parameters:
            values (int or array):
                Mask values. If ``None``, apply to entire
                ``Maskbit.mask`` array.  Default is ``None``.

        Returns:
            list:
                Bits that are set.

        Example:
            >>> maps = Maps(plateifu='8485-1901')
            >>> ha = maps['emline_gflux_ha_6564']
            >>> ha.pixmask.values_to_bits()
            [[[0, 1, 4, 30],
              [0, 1, 4, 30],
              ...
              [0, 1, 4, 30]]]
        """
        # assert (self.mask is not None) or (values is not None), 'Must provide values.'

        # values = np.array(self.mask) if values is None else np.array(values)
        # ndim = values.ndim

        # assert ndim <= 3, '`value` must be int, 1-D array, 2-D array, or 3-D array.'

        # # expand up to 2 dimensions
        # while values.ndim < 3:
        #     values = np.array([values])

        # # create list of list of lists of bits set
        # bits_set = []
        # for ii in range(values.shape[0]):
        #     row_ii = []
        #     for jj in range(values.shape[1]):
        #         row_jj = []
        #         for kk in range(values.shape[2]):
        #             row_jj.append(self._value_to_bits(values[ii, jj, kk], self.schema.bit.values))
        #         row_ii.append(row_jj)
        #     bits_set.append(row_ii)

        # # condense back down to initial dimensions
        # for __ in range(3 - ndim):
        #     bits_set = bits_set[0]

        bits_set = self._get_a_set(values, convert_to='bits')

        return bits_set

    def _get_uniq_bits(self, values):
        ''' Return a dictionary of unique bits

        Parameters:
            values (list):
                A flattened list of mask values

        Returns:
            dict:
                A unique dictionary of {mask value: bit list} as {key: value}
        '''
        uniqvals = set(values)
        vdict = {v: self._value_to_bits(v, self.schema.bit.values) for v in uniqvals}
        return vdict

    def _get_uniq_labels(self, values):
        ''' Return a dictionary of unique labels

        Parameters:
            values (list):
                A flattened list of mask values

        Returns:
            dict:
                A unique dictionary of {mask value: labels list} as {key: value}
        '''
        uniqbits = self._get_uniq_bits(values)
        uniqlabels = {k: self.schema.label[self.schema.bit.isin(v)].values.tolist() for k, v in uniqbits.items()}
        return uniqlabels

    def _get_a_set(self, values, convert_to='bits'):
        ''' Convert mask values to a list of either bit or label sets.

        Parameters:
            values (int or array):
                Mask values. If ``None``, apply to entire
                ``Maskbit.mask`` array.  Default is ``None``.
            convert_to (str):
                Indicates what to convert to.  Either "bits" or "labels"

        Returns:
            list:
                Bits/Labels that are set.

        '''
        assert (self.mask is not None) or (values is not None), 'Must provide values.'

        values = np.array(self.mask) if values is None else np.array(values)
        ndim = values.ndim
        shape = values.shape

        assert ndim <= 3, '`value` must be int, 1-D array, 2-D array, or 3-D array.'

        flatmask = values.flatten()

        if convert_to == 'bits':
            uniqvals = self._get_uniq_bits(flatmask)
        elif convert_to == 'labels':
            uniqvals = self._get_uniq_labels(flatmask)

        vallist = list(map(lambda x: uniqvals[x], flatmask))
        if ndim > 0:
            vals_set = np.reshape(vallist, shape).tolist()
        else:
            vals_set = vallist[0]

        return vals_set

    def _value_to_bits(self, value, bits_all):
        """Convert mask value to a list of bits.

        Parameters:
            value (int):
                Mask value.
            bits_all (array):
                All bits for flag.

        Returns:
            list:
                Bits that are set.
        """
        return [it for it in bits_all if int(value) & (1 << it)]

    def values_to_labels(self, values=None):
        """Convert mask values to a list of the labels of bits set.

        Parameters:
            values (int or array):
                Mask values. If ``None``, apply to entire
                ``Maskbit.mask`` array.  Default is ``None``.

        Returns:
            list:
                Bits that are set.

        Example:
            >>> maps = Maps(plateifu='8485-1901')
            >>> ha = maps['emline_gflux_ha_6564']
            >>> ha.pixmask.values_to_labels()
            [[['NOCOV', 'LOWCOV', 'NOVALUE', 'DONOTUSE'],
              ['NOCOV', 'LOWCOV', 'NOVALUE', 'DONOTUSE'],
               ...
              ['NOCOV', 'LOWCOV', 'NOVALUE', 'DONOTUSE']]]
        """
        #bits_set = self.values_to_bits(values=values)
        #labels_set = self._bits_to_labels(bits_set)

        labels_set = self._get_a_set(values, convert_to='labels')

        return labels_set

    def _bits_to_labels(self, nested):
        """Recursively convert a nested list of bits to labels.

        Parameters:
            nested (list):
                Nested list of bits.

        Returns:
            list: Nested list of labels.
        """
        # Base condition
        if isinstance(nested, (int, np.integer)):
            return self.schema.label[self.schema.bit == nested].values[0]

        return [self._bits_to_labels(it) for it in nested]

    def labels_to_value(self, labels):
        """Convert bit labels into a bit value.

        Parameters:
            labels (str or list):
                Labels of bits to set.

        Returns:
            int: Integer bit value.

        Example:
            >>> maps = Maps(plateifu='8485-1901')
            >>> ha = maps['emline_gflux_ha_6564']
            >>> ha.pixmask._labels_to_value('DONOTUSE')
            1073741824

            >>> ha.pixmask._labels_to_value(['NOCOV', 'LOWCOV'])
            3
        """
        if isinstance(labels, str):
            labels = [labels]

        bit_values = []
        for label in labels:
            bit = self.schema.bit[self.schema.label == label]
            if not bit.empty:
                bit_values.append(bit.values[0])

        return np.sum([2**value for value in bit_values])

    def labels_to_bits(self, labels):
        """Convert bit labels into bits.

        Parameters:
            labels (str or list):
                Labels of bits.

        Returns:
            list: Bits that correspond to the labels.

        Example:
            >>> maps = Maps(plateifu='8485-1901')
            >>> ha = maps['emline_gflux_ha_6564']
            >>> ha.pixmask.labels_to_bits('DONOTUSE')
            [30]

            >>> ha.pixmask.labels_to_value(['NOCOV', 'LOWCOV'])
            [0, 1]
        """
        return self.values_to_bits(self.labels_to_value(labels))

    def get_mask(self, labels, mask=None, dtype=int):
        """Create mask from a list of labels.

        If ``dtype`` is ``int``, then ``get_mask`` can effectively
        perform an OR or AND operation.  However, if ``dtype`` is
        ``bool``, then ``get_mask`` does an OR.

        Parameters:
            labels (str or list):
                Labels of bits.
            mask (int or array):
                User-defined mask. If ``None``, use ``self.mask``.
                Default is ``None``.
            dtype:
                Output dtype, which must be either ``int`` or ``bool``.
                Default is ``int``.

        Returns:
            array: Mask for given labels.

        Example:
            >>> maps = Maps(plateifu='8485-1901')
            >>> ha = maps['emline_gflux_ha_6564']
            >>> ha.pixmask.get_mask(['NOCOV', 'LOWCOV'])
            array([[3, 3, 3, ..., 3, 3, 3],
                   ...,
                   [3, 3, 3, ..., 3, 3, 3]])

            >>> ha.pixmask.get_mask(['NOCOV', 'LOWCOV'], dtype=bool)
            array([[ True,  True,  True, ...,  True,  True,  True],
                   ...,
                   [ True,  True,  True, ...,  True,  True,  True]], dtype=bool)
        """

        assert dtype in [int, bool], '``dtype`` must be either ``int`` or ``bool``.'

        if isinstance(labels, str):
            labels = [labels]

        schema_labels = self.schema.label.tolist()
        for label in labels:
            if label not in schema_labels:
                raise ValueError('label {0!r} not found in the maskbit schema.'.format(label))

        bits = self.labels_to_bits(labels)
        mask = mask if mask is not None else self.mask

        if len(bits) == 0:
            return np.zeros(mask.shape, dtype=np.int)

        return np.sum([mask & 2**bit for bit in bits], axis=0).astype(dtype)



#!/usr/bin/env python
# encoding: utf-8

# Licensed under a 3-clause BSD license.
#

from __future__ import print_function

import copy
import datetime
import json
import os
import warnings
from collections import namedtuple
from functools import wraps
from operator import add

import numpy as np
import six
from astropy.table import Table, hstack, vstack
from fuzzywuzzy import process
import marvin.utils.plot.scatter
from marvin import config, log
from marvin.api.api import Interaction
from marvin.core import marvin_pickle
from marvin.core.exceptions import (MarvinBreadCrumb, MarvinError, MarvinUserWarning)
from marvin.tools.cube import Cube
from marvin.tools.maps import Maps
from marvin.tools.modelcube import ModelCube
from marvin.tools.rss import RSS
from marvin.utils.datamodel.query import datamodel
from marvin.utils.datamodel.query.base import ParameterGroup
from marvin.utils.general import (downloadList, get_images_by_list, map_bins_to_column, temp_setattr,
                                  turn_off_ion)

try:
    import cPickle as pickle
except:
    import pickle

try:
    import pandas as pd
except ImportError:
    pd = None
    warnings.warn('Could not import pandas.', MarvinUserWarning)

__all__ = ['Results', 'ResultSet']

breadcrumb = MarvinBreadCrumb()


def local_mode_only(fxn):
    '''Decorator that bypasses function if in remote mode.'''

    @wraps(fxn)
    def wrapper(self, *args, **kwargs):
        if self.mode == 'remote':
            raise MarvinError('{0} not available in remote mode'.format(fxn.__name__))
        else:
            return fxn(self, *args, **kwargs)
    return wrapper


def remote_mode_only(fxn):
    '''Decorator that bypasses function if in local mode.'''

    @wraps(fxn)
    def wrapper(self, *args, **kwargs):
        if self.mode == 'local':
            raise MarvinError('{0} not available in local mode'.format(fxn.__name__))
        else:
            return fxn(self, *args, **kwargs)
    return wrapper


class ColumnGroup(ParameterGroup):
    ''' Subclass of Parameter Group '''

    def __repr__(self):
        ''' New repr for the Results Column Parameter Group '''
        old = list.__repr__(self)
        old = old.replace('>,', '>,\n')
        return ('<ParameterGroup name={0.name}, n_parameters={1}>\n '
                '{2}'.format(self, len(self), old))

    def __str__(self):
        ''' New string repr for prints '''
        return self.__repr__()


def marvintuple(name, params=None, **kwargs):
    ''' Custom namedtuple class factory for Marvin Results rows

    A class factory designed to create a new Marvin ResultRow class object. marvintuple
    creates a new class definition which can be instantiated.  Parameters can be pushed into
    an instance as individual arguments, or as key-value pairs.  See the
    `namedtuple <https://docs.python.org/2/library/collections.html#collections.namedtuple>`_
    for details on the Python container.

    Parameters:
        name (str):
            The name of the Class.  Required.
        params (str|list):
            The list of parameters to add as fields to the namedtuple.  Can be a list of names
            or a comma-separated string of names.

    Returns:
        a new namedtuple class

    Example:
        >>> # create new class with two fields
        >>> mt = marvintuple('Row', ['mangaid', 'plateifu'])
        >>>
        >>> # create a new instance of the class, with values
        >>> row = mt('1-209232', '8485-1901')
        >>> # or
        >>> row = mt(mangaid='1-209232', plateifu='8485-1901')

    '''

    # check the params input
    if params and isinstance(params, six.string_types):
        params = params.split(',') if ',' in params else [params]
        params = [p.strip() for p in params]

    # pop any extra keywords
    results = kwargs.pop('results', None)

    # create default namedtuple
    nt = namedtuple(name, params, **kwargs)

    def new_add(self, other):
        ''' Overloaded add to combine tuples without duplicates '''

        if self.release:
            assert self.release == other.release, 'Cannot add result rows from different releases'
        if self._search_filter:
            assert self._search_filter == other._search_filter, ('Cannot add result rows generated '
                                                                 'using different search filters')

        assert hasattr(self, 'plateifu') and hasattr(other, 'plateifu'), ("All rows must have a "
                                                                          "plateifu column to be able to add")

        assert self.plateifu == other.plateifu, 'The plateifus must be the same to add these rows'

        self_dict = self._asdict()
        other_dict = other._asdict()
        self_dict.update(other_dict)

        new_fields = tuple(self_dict.keys())
        marvin_row = marvintuple(self.__class__.__name__, new_fields, results=self._results)
        return marvin_row(**self_dict)

    # append new properties and overloaded methods
    nt.__add__ = new_add
    nt._results = results
    nt.release = results.release if results else None
    nt._search_filter = results.search_filter if results else None
    return nt


class ResultSet(list):
    ''' A Set of Results

    A list object representing a set of query results.  Each row of the list is a
    ResultRow object, which is a custom Marvin namedtuple object.  ResultSets can be
    extended column-wise or row-wise by adding them together.

    Parameters:
        _objects (list):
            A list of objects. Required.
        count (int):
            The count of objects in the current list
        totalcount (int):
            The total count of objects in the full results
        index (int):
            The index of the current set within the total set.
        columns (list):
            A list of columns accompanying this set
        results (Results):
            The Marvin Results object this set is a part of

    '''
    def __init__(self, _objects, **kwargs):
        list.__init__(self, _objects)
        self._results = kwargs.get('results', None)
        self.columns = kwargs.get('columns', None)
        self.count = kwargs.get('count', None)
        self.total = kwargs.get('total', None)
        self._populate_from_results()
        self.pages = int(np.ceil(self.total / float(self.count))) if self.count else 0
        self.index = kwargs.get('index') if kwargs.get('index') else 0
        self.end_index = self.index + self.count
        self.current_page = (int(self.index) + self.count) / self.count

    def __repr__(self):
        old = list.__repr__(self)
        return ('<ResultSet(set={0.current_page}/{0.pages}, index={0.index}:{0.end_index}, '
                'count_in_set={0.count}, total={0.total})>\n{1}'.format(self, old.replace('),', '),\n')))

    def __getitem__(self, value):
        if isinstance(value, six.string_types):
            value = str(value)
            if value in self.columns:
                colname = self.columns[value].remote
                rows = [row.__getattribute__(colname) for row in self]
            else:
                rows = [row for row in self if value in row]
            if rows:
                return rows[0] if len(rows) == 1 else rows
            else:
                raise ValueError('{0} not found in the list'.format(value))
        elif isinstance(value, int):
            return list.__getitem__(self, value)
        elif isinstance(value, slice):
            newset = list.__getitem__(self, value)
            return ResultSet(newset, index=int(value.start), count=len(newset), total=self.total, columns=self.columns, results=self._results)
        elif isinstance(value, np.ndarray):
            return np.array(self)[value]

    def __getslice__(self, start, stop):
        newset = list.__getslice__(self, start, stop)
        return ResultSet(newset, index=start, count=len(newset), total=self.total, columns=self.columns, results=self._results)

    def __add__(self, other):
        newresults = self._results

        if not isinstance(other, ResultSet):
            raise MarvinUserWarning('Can only add ResultSets together')

        # add elements
        if self.index == other.index:
            # column-wise add
            newcols = self.columns.full + [col.full for col in other.columns if col.full not in self.columns.full]
            parent = self._results.datamodel if self._results else None
            newcols = ColumnGroup('Columns', newcols, parent=parent)
            newresults.columns = newcols
            new_set = map(add, self, other)
        else:
            # row-wise add

            # warn if the subsets are not consecutive
            if abs(self.index - other.index) > self.count:
                warnings.warn('You are combining non-consectuive sets! '
                              'The indexing and ordering will be messed up')

            # copy the sets
            new_set = copy.copy(self) if self.index < other.index else copy.copy(other)
            set_b = copy.copy(other) if self.index < other.index else copy.copy(self)
            # filter out any rows that already exist in the set
            rows = [row for row in set_b if row not in new_set]
            # extend the set
            new_set.extend(rows)
            newcols = self.columns
            self.count = len(new_set)
            self.index = min(self.index, other.index)

        return ResultSet(new_set, count=self.count, total=self.total, index=self.index,
                         columns=newcols, results=newresults)

    def __radd__(self, other):
        return self.__add__(other)

    def _populate_from_results(self):
        ''' Populate some parameters from the results '''
        if self._results:
            self.columns = self._results.columns if not self.columns else self.columns
            self.choices = self.columns.list_params('remote')
            self.count = self._results.count if not self.count else self.count
            self.total = self._results.totalcount if not self.total else self.total
        else:
            self.count = self.count if self.count else len(self)
            self.total = self.total if self.total else len(self)

    def to_dict(self, name=None, format_type='listdict'):
        ''' Convert the ResultSet into a dictionary

        Converts the set of results into a list of dictionaries.  Optionally
        accepts a column name keyword to extract only that column.

        Parameters:
            name (str):
                Name of the column you wish to extract.  Default is None.
            format_type (str):
                The format of the output dictionary.  Can either be a list of dictionaries
                or a dictionary of lists.

        Returns:
            The output converted into dictionary format.
        '''

        keys = self.columns.list_params('remote')

        if format_type == 'listdict':
            if name:
                output = [{k: res.__getattribute__(k) for k in [name]} for res in self]
            else:
                output = [{k: res.__getattribute__(k) for k in keys} for res in self]
        elif format_type == 'dictlist':
            if name:
                output = {k: [res._asdict()[k] for res in self] for k in [name]}
            else:
                output = {k: [res._asdict()[k] for res in self] for k in keys}
        else:
            raise MarvinError('Cannot output dictionaries.  Check your input format_type.')
        return output

    def to_list(self):
        ''' Converts to a standard Python list object '''
        return list(self)

    def sort(self, name=None, reverse=False):
        ''' Sort the results

        In-place sorting of the result set.  This is the standard list sorting mechanism.
        When no name is specified, does standard list sorting with no key.

        Parameters:
            name (str):
                Column name to sort on.  Default is None.
            reverse (bool):
                If True, sorts in reverse (descending) order.

        Returns:
            A sorted list

        '''
        if name:
            colname = self.columns[name].remote
            return list.sort(self, key=lambda row: row.__getattribute__(colname), reverse=reverse)
        else:
            return list.sort(self)


class Results(object):
    ''' A class to handle results from queries on the MaNGA dataset

    Parameters:
        results (list):
            List of results satisfying the input Query
        query (object / str):
            The query used to produce these results. In local mode, the query is an
            SQLalchemy object that can be used to redo the query, or extract subsets
            of results from the query. In remote more, the query is a literal string
            representation of the SQL query.
        return_type (str):
            The MarvinTools object to convert the results into.  If initially set, the results
            are automaticaly converted into the specified Marvin Tool Object on initialization
        objects (list):
            The list of Marvin Tools objects created by returntype
        count (int):
            The number of objects in the returned query results
        totalcount (int):
            The total number of objects in the full query results
        mode ({'auto', 'local', 'remote'}):
            The load mode to use. See :doc:`Mode secision tree</mode_decision>`.
        chunk (int):
            For paginated results, the number of results to return.  Defaults to 10.
        start (int):
            For paginated results, the starting index value of the results.  Defaults to 0.
        end (int):
            For paginated results, the ending index value of the resutls.  Defaults to start+chunk.

    Attributes:
        count (int):  The count of objects in your current page of results
        totalcount (int): The total number of results in the query
        query_time (datetime): A datetime TimeDelta representation of the query runtime

    Returns:
        results: An object representing the Results entity

    Example:
        >>> f = 'nsa.z < 0.012 and ifu.name = 19*'
        >>> q = Query(search_filter=f)
        >>> r = q.run()
        >>> print(r)
        >>> Results(results=[(u'4-3602', u'1902', -9999.0), (u'4-3862', u'1902', -9999.0), (u'4-3293', u'1901', -9999.0), (u'4-3988', u'1901', -9999.0), (u'4-4602', u'1901', -9999.0)],
        >>>         query=<sqlalchemy.orm.query.Query object at 0x115217090>,
        >>>         count=64,
        >>>         mode=local)

    '''

    def __init__(self, results=None, mode=None, data_origin=None, release=None, count=None,
                 totalcount=None, runtime=None, response_time=None, chunk=None, start=None,
                 end=None, queryobj=None, query=None, search_filter=None, return_params=None,
                 return_type=None, limit=None, params=None, **kwargs):

        # basic parameters
        self.results = results
        self.mode = mode if mode else config.mode
        self.data_origin = data_origin
        self.objects = None

        # input query parameters
        self._queryobj = queryobj
        self._params = self._queryobj.params if self._queryobj else params
        self.release = self._queryobj.release if self._queryobj else release
        self.query = self._queryobj.query if self._queryobj else query
        self.return_type = self._queryobj.return_type if self._queryobj else return_type
        self.search_filter = self._queryobj.search_filter if self._queryobj else search_filter
        self.return_params = self._queryobj.return_params if self._queryobj else return_params
        self.limit = self._queryobj.limit if self._queryobj else limit

        # stat parameters
        self.datamodel = datamodel[self.release]
        self.count = count if count else len(self.results)
        self.totalcount = totalcount if totalcount else self.count
        self._runtime = runtime
        self.query_time = self._getRunTime() if self._runtime is not None else None
        self.response_time = response_time

        # ordering parameters
        self.chunk = chunk
        self.start = start
        self.end = end
        self.sortcol = None
        self.order = None

        # drop breadcrumb
        breadcrumb.drop(message='Initializing MarvinResults {0}'.format(self.__class__),
                        category=self.__class__)

        # Convert results to MarvinTuple
        if self.count > 0 and self.results:
            self._set_page()
            self._create_result_set(index=self.start)

        # Auto convert to Marvin Object
        if self.return_type:
            self.convertToTool(self.return_type)

    def __add__(self, other):
        assert isinstance(other, Results) is True, 'Can only add Marvin Results together'
        assert self.release == other.release, 'Cannot add Marvin Results from different releases'
        assert self.search_filter == other.search_filter, 'Cannot add Marvin Results with different search filters'
        results = self.results + other.results
        return_params = self.return_params + [p for p in other.return_params if p not in self.return_params]
        params = self._params + [p for p in other._params if p not in self._params]
        return Results(results=results, params=params, return_params=return_params, limit=self.limit,
                       search_filter=self.search_filter, count=len(results), totalcount=self.totalcount,
                       release=self.release, mode=self.mode)

    def __radd__(self, other):
        return self.__add__(other)

    def __repr__(self):
        return ('Marvin Results(query={0}, totalcount={1}, count={2}, mode={3})'.format(self.search_filter, self.totalcount, self.count, self.mode))

    def showQuery(self):
        ''' Displays the literal SQL query used to generate the Results objects

        Returns:
            querystring (str):
                A string representation of the SQL query
        '''

        # check unicode or str
        isstr = isinstance(self.query, six.string_types)

        # return the string query or compile the real query
        if isstr:
            return self.query
        else:
            return str(self.query.statement.compile(compile_kwargs={'literal_binds': True}))

    def _getRunTime(self):
        ''' Sets the query runtime as a datetime timedelta object '''
        if isinstance(self._runtime, dict):
            return datetime.timedelta(**self._runtime)
        else:
            return self._runtime

    def download(self, images=False, limit=None):
        ''' Download results via sdss_access

        Uses sdss_access to download the query results via rsync.
        Downloads them to the local sas. The data type downloaded
        is indicated by the returntype parameter

        i.e. $SAS_BASE_DIR/mangawork/manga/spectro/redux/...

        Parameters:
            images (bool):
                Set to only download the images of the query results
            limit (int):
                A limit of the number of results to download

        Returns:
            NA: na

        Example:
            >>> r = q.run()
            >>> r.returntype = 'cube'
            >>> r.download()
        '''

        plateifu = self.getListOf('plateifu')
        if images:
            tmp = get_images_by_list(plateifu, releas=self.release, download=True)
        else:
            downloadList(plateifu, dltype=self.return_type, limit=limit)

    def sort(self, name, order='asc'):
        ''' Sort the set of results by column name

        Sorts the results (in place) by a given parameter / column name.  Sets
        the results to the new sorted results.

        Parameters:
            name (str):
                The column name to sort on
            order ({'asc', 'desc'}):
                To sort in ascending or descending order.  Default is asc.

        Example:
            >>> r = q.run()
            >>> r.getColumns()
            >>> [u'mangaid', u'name', u'nsa.z']
            >>> r.results
            >>> [(u'4-3988', u'1901', -9999.0),
            >>>  (u'4-3862', u'1902', -9999.0),
            >>>  (u'4-3293', u'1901', -9999.0),
            >>>  (u'4-3602', u'1902', -9999.0),
            >>>  (u'4-4602', u'1901', -9999.0)]

            >>> # Sort the results by mangaid
            >>> r.sort('mangaid')
            >>> [(u'4-3293', u'1901', -9999.0),
            >>>  (u'4-3602', u'1902', -9999.0),
            >>>  (u'4-3862', u'1902', -9999.0),
            >>>  (u'4-3988', u'1901', -9999.0),
            >>>  (u'4-4602', u'1901', -9999.0)]

            >>> # Sort the results by IFU name in descending order
            >>> r.sort('ifu.name', order='desc')
            >>> [(u'4-3602', u'1902', -9999.0),
            >>>  (u'4-3862', u'1902', -9999.0),
            >>>  (u'4-3293', u'1901', -9999.0),
            >>>  (u'4-3988', u'1901', -9999.0),
            >>>  (u'4-4602', u'1901', -9999.0)]
        '''
        remotename = self._check_column(name, 'remote')
        self.sortcol = remotename
        self.order = order

        if self.mode == 'local':
            reverse = True if order == 'desc' else False
            self.getAll()
            self.results.sort(remotename, reverse=reverse)
            self.results = self.results[0:self.limit]
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['querycubes']['url']
            params = {'searchfilter': self.search_filter, 'returnparams': self.return_params,
                      'sort': remotename, 'order': order, 'limit': self.limit}

            self._interaction(url, params, create_set=True, calltype='Sort')

        return self.results

    def toTable(self):
        ''' Output the results as an Astropy Table

        Uses the Python Astropy package

        Parameters:
            None

        Returns:
            tableres:
                Astropy Table

        Example:
            >>> r = q.run()
            >>> r.toTable()
            >>> <Table length=5>
            >>> mangaid    name   nsa.z
            >>> unicode6 unicode4   float64
            >>> -------- -------- ------------
            >>>   4-3602     1902      -9999.0
            >>>   4-3862     1902      -9999.0
            >>>   4-3293     1901      -9999.0
            >>>   4-3988     1901      -9999.0
            >>>   4-4602     1901      -9999.0
        '''
        try:
            tabres = Table(rows=self.results, names=self.columns.full)
        except ValueError as e:
            raise MarvinError('Could not make astropy Table from results: {0}'.format(e))
        return tabres

    def merge_tables(self, tables, direction='vert', **kwargs):
        ''' Merges a list of Astropy tables of results together

        Combines two Astropy tables using either the Astropy
        vstack or hstack method.  vstack refers to vertical stacking of table rows.
        hstack refers to horizonal stacking of table columns.  hstack assumes the rows in each
        table refer to the same object.  Buyer beware: stacking tables without proper understanding
        of your rows and columns may results in deleterious results.

        merge_tables also accepts all keyword arguments that Astropy vstack and hstack method do.
        See `vstack <http://docs.astropy.org/en/stable/table/operations.html#stack-vertically>`_
        See `hstack <http://docs.astropy.org/en/stable/table/operations.html#stack-horizontally>`_

        Parameters:
            tables (list):
                A list of Astropy Table objects.  Required.
            direction (str):
                The direction of the table stacking, either vertical ('vert') or horizontal ('hor').
                Default is 'vert'.  Direction string can be fuzzy.

        Returns:
            A new Astropy table that is the stacked combination of all input tables

        Example:
            >>> # query 1
            >>> q, r = doQuery(search_filter='nsa.z < 0.1', returnparams=['g_r', 'cube.ra', 'cube.dec'])
            >>> # query 2
            >>> q2, r2 = doQuery(search_filter='nsa.z < 0.1')
            >>>
            >>> # convert to tables
            >>> table_1 = r.toTable()
            >>> table_2 = r2.toTable()
            >>> tables = [table_1, table_2]
            >>>
            >>> # vertical (row) stacking
            >>> r.merge_tables(tables, direction='vert')
            >>> # horizontal (column) stacking
            >>> r.merge_tables(tables, direction='hor')

        '''
        choices = ['vertical', 'horizontal']
        stackdir, score = process.extractOne(direction, choices)
        if stackdir == 'vertical':
            return vstack(tables, **kwargs)
        elif stackdir == 'horizontal':
            return hstack(tables, **kwargs)

    def toFits(self, filename='myresults.fits', overwrite=False):
        ''' Output the results as a FITS file

        Writes a new FITS file from search results using
        the astropy Table.write()

        Parameters:
            filename (str):
                Name of FITS file to output
            overwrite (bool):
                Set to True to overwrite an existing file

        '''
        myext = os.path.splitext(filename)[1]
        if not myext:
            filename = filename + '.fits'
        table = self.toTable()
        table.write(filename, format='fits', overwrite=overwrite)
        print('Writing new FITS file {0}'.format(filename))

    def toCSV(self, filename='myresults.csv', overwrite=False):
        ''' Output the results as a CSV file

        Writes a new CSV file from search results using
        the astropy Table.write()

        Parameters:
            filename (str):
                Name of CSV file to output
            overwrite (bool):
                Set to True to overwrite an existing file

        '''
        myext = os.path.splitext(filename)[1]
        if not myext:
            filename = filename + '.csv'
        table = self.toTable()
        table.write(filename, format='csv', overwrite=overwrite)
        print('Writing new CSV file {0}'.format(filename))

    def toDF(self):
        '''Call toDataFrame().
        '''
        return self.toDataFrame()

    def toDataFrame(self):
        '''Output the results as an pandas dataframe.

        Uses the pandas package.

        Parameters:
            None

        Returns:
            dfres:
                pandas dataframe

        Example:
            >>> r = q.run()
            >>> r.toDataFrame()
            mangaid  plate   name     nsa_mstar         z
            0  1-22286   7992  12704  1.702470e+11  0.099954
            1  1-22301   7992   6101  9.369260e+10  0.105153
            2  1-22414   7992   6103  7.489660e+10  0.092272
            3  1-22942   7992  12705  8.470360e+10  0.104958
            4  1-22948   7992   9102  1.023530e+11  0.119399
        '''
        res = self.results.to_list() if self.results else []
        try:
            dfres = pd.DataFrame(res)
        except (ValueError, NameError) as e:
            raise MarvinError('Could not make pandas dataframe from results: {0}'.format(e))
        return dfres

    def _create_result_set(self, index=None, rows=None):
        ''' Creates a Marvin ResultSet

        Parameters:
            index (int):
                The starting index of the result subset
            rows (list|ResultSet):
                A list of rows containing the value data to input into the ResultSet

        Returns:
            creates a marvin ResultSet and sets it as the results attribute

        '''

        # grab the columns from the results
        self.columns = self.getColumns()
        ntnames = self.columns.list_params('remote')
        # dynamically create a new ResultRow Class
        rows = rows if rows else self.results
        row_is_dict = isinstance(rows[0], dict)
        if not isinstance(rows, ResultSet):
            nt = marvintuple('ResultRow', ntnames, results=self)
            if row_is_dict:
                results = [nt(**r) for r in rows]
            else:
                results = [nt(*r) for r in rows]
        else:
            results = rows
        self.count = len(results)
        # Build the ResultSet
        self.results = ResultSet(results, count=self.count, total=self.totalcount, index=index, results=self)

    def _set_page(self):
        ''' Set the page of the data '''
        if self.start and self.end:
            self.chunk = (self.end - self.start)
        else:
            self.chunk = self.chunk if self.chunk else self.limit if self.limit else 100
            self.start = 0
            self.end = self.start + self.chunk

        self.pages = int(np.ceil(self.totalcount / float(self.count))) if self.count else 0
        self.index = self.start
        self.current_page = (int(self.index) + self.count) / self.count

    def save(self, path=None, overwrite=False):
        ''' Save the results as a pickle object

        Parameters:
            path (str):
                Filepath and name of the pickled object
            overwrite (bool):
                Set this to overwrite an existing pickled file

        Returns:
            path (str):
                The filepath and name of the pickled object

        '''

        # set the filename and path
        sf = self.search_filter.replace(' ', '') if self.search_filter else 'anon'
        # set the path
        if not path:
            path = os.path.expanduser('~/marvin_results_{0}.mpf'.format(sf))

        # check for file extension
        if not os.path.splitext(path)[1]:
            path = os.path.join(path + '.mpf')

        path = os.path.realpath(path)

        if os.path.isdir(path):
            raise MarvinError('path must be a full route, including the filename.')

        if os.path.exists(path) and not overwrite:
            warnings.warn('file already exists. Not overwriting.', MarvinUserWarning)
            return

        dirname = os.path.dirname(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        # convert results into a dict
        dict_results = self.results.to_dict()

        # set bad pickled attributes to None
        attrs = ['results', 'datamodel', 'columns', '_queryobj']
        vals = [dict_results, None, None, None]
        isnotstr = not isinstance(self.query, six.string_types)
        if isnotstr:
            attrs += ['query']
            vals += [None]

        # pickle the results
        try:
            with temp_setattr(self, attrs, vals):
                pickle.dump(self, open(path, 'wb'), protocol=-1)
        except Exception as ee:
            if os.path.exists(path):
                os.remove(path)
            raise MarvinError('Error found while pickling: {0}'.format(str(ee)))

        return path

    @classmethod
    def restore(cls, path, delete=False):
        ''' Restore a pickled Results object

        Parameters:
            path (str):
                The filename and path to the pickled object

            delete (bool):
                Turn this on to delete the pickled fil upon restore

        Returns:
            Results (instance):
                The instantiated Marvin Results class
        '''
        obj = marvin_pickle.restore(path, delete=delete)
        obj.datamodel = datamodel[obj.release]
        obj._create_result_set()
        obj.getColumns()
        return obj

    def toJson(self, orient: str = 'records', pure: bool = None) -> str:
        ''' Output the results as a JSON object

        Uses Python panda package to convert the results to a JSON object.  The default 
        orientation is a list "records".  Valid orientations are ('split', 'records', 
        'index', 'columns', 'values', 'table').  If pandas is not installed or the "pure" 
        option is set, will use the json package to convert the results to JSON representation.

        Parameters:
            orient (str):
                The pandas orientation to use when converting to JSON. Default is 'records'.
            pure (bool):
                Set this to True to use the json library for conversion instead of the pandas package

        Returns:
            str:
                The results as a JSON string
        '''
        # if no pandas or pure is true, then use json dumps
        if not pd or pure:
            try:
                jsonres = json.dumps(self.results)
            except TypeError as e:
                raise MarvinError('Results not JSON-ifiable. Check the format of results: {0}'.format(e))
        else:
            jsonres = self.toDF().to_json(orient=orient)
        return jsonres

    def getColumns(self):
        ''' Get the columns of the returned reults

        Returns a ParameterGroup containing the columns from the
        returned results.  Each row of the ParameterGroup is a
        QueryParameter.

        Returns:
            columns (list):
                A list of column names from the results

        Example:
            >>> r = q.run()
            >>> cols = r.getColumns()
            >>> print(cols)
            >>> [u'mangaid', u'name', u'nsa.z']

        '''
        try:
            self.columns = ColumnGroup('Columns', self._params, parent=self.datamodel)
        except Exception as e:
            raise MarvinError('Could not create query columns: {0}'.format(e))
        return self.columns

    def _interaction(self, url, params, calltype='', create_set=None, **kwargs):
        ''' Perform a remote Interaction call

        Parameters:
            url (str):
                The url of the request
            params (dict):
                A dictionary of parameters (get or post) to send with the request
            calltype (str):
                The method call sending the request
            create_set (bool):
                If True, sets the response output as the new results and creates a
                new named tuple set

        Returns:
            output:
                The output data from the request

        Raises:
            MarvinError: Raises on any HTTP Request error

        '''

        # check if the returnparams parameter is in the proper format
        if 'returnparams' in params:
            return_params = params.get('returnparams', None)
            if return_params and isinstance(return_params, list):
                params['returnparams'] = ','.join(return_params)

        # add the release just in case
        params.update({'release': self.release})

        # check if we're getting all results
        datastream = calltype == 'getAll'

        # send the request
        try:
            ii = Interaction(route=url, params=params, stream=True, datastream=datastream)
        except MarvinError as e:
            raise MarvinError('API Query {0} call failed: {1}'.format(calltype, e))
        else:
            remotes = self._queryobj._get_remote_parameters(ii)
            output = remotes['results']
            self.response_time = remotes['response_time']
            self._runtime = remotes['runtime']
            self.query_time = self._getRunTime()
            index = kwargs.get('index', None)
            if create_set:
                self._create_result_set(index=index, rows=output)
            else:
                return output

    def _check_column(self, name, name_type):
        ''' Check if a name exists as a column '''
        try:
            name_in_col = name in self.columns
        except KeyError as e:
            raise MarvinError('Column {0} not found in results: {1}'.format(name, e))
        else:
            assert name_type in ['full', 'remote', 'name', 'short', 'display'], \
                'name_type must be one of "full, remote, name, short, display"'
            return self.columns[str(name)].__getattribute__(name_type)

    def getListOf(self, name=None, to_json=False, to_ndarray=False, return_all=None):
        ''' Extract a list of a single parameter from results

        Parameters:
            name (str):
                Name of the parameter name to return.  If not specified,
                it returns all parameters.
            to_json (bool):
                True/False boolean to convert the output into a JSON format
            to_ndarray (bool):
                True/False boolean to convert the output into a Numpy array
            return_all (bool):
                if True, returns the entire result set for that column

        Returns:
            output (list):
                A list of results for one parameter

        Example:
            >>> r = q.run()
            >>> r.getListOf('mangaid')
            >>> [u'4-3988', u'4-3862', u'4-3293', u'4-3602', u'4-4602']

        Raises:
            AssertionError:
                Raised when no name is specified.
        '''

        assert name, 'Must specify a column name'

        # check column name and get full name
        fullname = self._check_column(name, 'full')

        # deal with the output
        if return_all:
            # # grab all of that column
            url = config.urlmap['api']['getcolumn']['url'].format(colname=fullname)
            params = {'searchfilter': self.search_filter, 'format_type': 'list',
                      'return_all': True, 'returnparams': self.return_params}
            output = self._interaction(url, params, calltype='getList')
        else:
            # only deal with current page
            output = self.results[name] if self.results.count > 1 else [self.results[name]]

        if to_json:
            output = json.dumps(output) if output else None

        if to_ndarray:
            output = np.array(output) if output else None

        return output

    def getDictOf(self, name=None, format_type='listdict', to_json=False, return_all=None):
        ''' Get a dictionary of specified parameters

        Parameters:
            name (str):
                Name of the parameter name to return.  If not specified,
                it returns all parameters.
            format_type ({'listdict', 'dictlist'}):
                The format of the results. Listdict is a list of dictionaries.
                Dictlist is a dictionary of lists. Default is listdict.
            to_json (bool):
                True/False boolean to convert the output into a JSON format
            return_all (bool):
                if True, returns the entire result set for that column

        Returns:
            output (list, dict):
                Can be either a list of dictionaries, or a dictionary of lists

        Example:
            >>> # get some results
            >>> r = q.run()
            >>> # Get a list of dictionaries
            >>> r.getDictOf(format_type='listdict')
            >>> [{'cube.mangaid': u'4-3988', 'ifu.name': u'1901', 'nsa.z': -9999.0},
            >>>  {'cube.mangaid': u'4-3862', 'ifu.name': u'1902', 'nsa.z': -9999.0},
            >>>  {'cube.mangaid': u'4-3293', 'ifu.name': u'1901', 'nsa.z': -9999.0},
            >>>  {'cube.mangaid': u'4-3602', 'ifu.name': u'1902', 'nsa.z': -9999.0},
            >>>  {'cube.mangaid': u'4-4602', 'ifu.name': u'1901', 'nsa.z': -9999.0}]

            >>> # Get a dictionary of lists
            >>> r.getDictOf(format_type='dictlist')
            >>> {'cube.mangaid': [u'4-3988', u'4-3862', u'4-3293', u'4-3602', u'4-4602'],
            >>>  'ifu.name': [u'1901', u'1902', u'1901', u'1902', u'1901'],
            >>>  'nsa.z': [-9999.0, -9999.0, -9999.0, -9999.0, -9999.0]}

            >>> # Get a dictionary of only one parameter
            >>> r.getDictOf('mangaid')
            >>> [{'cube.mangaid': u'4-3988'},
            >>>  {'cube.mangaid': u'4-3862'},
            >>>  {'cube.mangaid': u'4-3293'},
            >>>  {'cube.mangaid': u'4-3602'},
            >>>  {'cube.mangaid': u'4-4602'}]

        '''

        # get the remote and full name
        remotename = self._check_column(name, 'remote') if name else None
        fullname = self._check_column(name, 'full') if name else None

        # create the dictionary
        output = self.results.to_dict(name=remotename, format_type=format_type)

        # deal with the output
        if return_all:
            # grab all or of a specific column
            params = {'searchfilter': self.search_filter, 'return_all': True,
                      'format_type': format_type, 'returnparams': self.return_params}
            url = config.urlmap['api']['getcolumn']['url'].format(colname=fullname)
            output = self._interaction(url, params, calltype='getDict')
        else:
            # only deal with current page
            output = self.results.to_dict(name=remotename, format_type=format_type)

        if to_json:
            output = json.dumps(output) if output else None

        return output

    def loop(self, chunk=None):
        ''' Loop over the full set of results

        Starts a loop to collect all the results (in chunks)
        until the current count reaches the total number
        of results.  Uses extendSet.

        Parameters:
            chunk (int):
                The number of objects to return

        Example:
            >>> # get some results from a query
            >>> r = q.run()
            >>> # start a loop, grabbing in chunks of 400
            >>> r.loop(chunk=400)

        '''

        while self.count < self.totalcount:
            self.extendSet(chunk=chunk)

    def extendSet(self, chunk=None, start=None):
        ''' Extend the Result set with the next page

        Extends the current ResultSet with the next page of results
        or a specified page.  Calls either getNext or getSubset.

        Parameters:
            chunk (int):
                The number of objects to return
            start (int):
                The starting index of your subset extraction

        Returns:
            A new results set

        Example:
            >>> # run a query
            >>> r = q.run()
            >>> # extend the current result set with the next page
            >>> r.extendSet()
            >>>

        See Also:
            getNext, getSubset

        '''

        oldset = copy.copy(self.results)
        if start is not None:
            nextset = self.getSubset(start, limit=chunk)
        else:
            nextset = self.getNext(chunk=chunk)
        newset = oldset + nextset
        self.count = len(newset)
        self.results = newset

    def getNext(self, chunk=None):
        ''' Retrieve the next chunk of results

        Returns the next chunk of results from the query.
        from start to end in units of chunk.  Used with getPrevious
        to paginate through a long list of results

        Parameters:
            chunk (int):
                The number of objects to return

        Returns:
            results (list):
                A list of query results

        Example:
            >>> r = q.run()
            >>> r.getNext(5)
            >>> Retrieving next 5, from 35 to 40
            >>> [(u'4-4231', u'1902', -9999.0),
            >>>  (u'4-14340', u'1901', -9999.0),
            >>>  (u'4-14510', u'1902', -9999.0),
            >>>  (u'4-13634', u'1901', -9999.0),
            >>>  (u'4-13538', u'1902', -9999.0)]

        See Also:
            getAll, getPrevious, getSubset

        '''

        if chunk and chunk < 0:
            warnings.warn('Chunk cannot be negative. Setting to {0}'.format(self.chunk), MarvinUserWarning)
            chunk = self.chunk

        newstart = self.end
        self.chunk = chunk if chunk else self.chunk
        newend = newstart + self.chunk

        # This handles cases when the number of results is < total
        if self.totalcount == self.count:
            warnings.warn('You have all the results.  Cannot go forward', MarvinUserWarning)
            return self.results

        # This handles the end edge case
        if newend > self.totalcount:
            warnings.warn('You have reached the end.', MarvinUserWarning)
            newend = self.totalcount
            newstart = self.end

        # This grabs the next chunk
        log.info('Retrieving next {0}, from {1} to {2}'.format(self.chunk, newstart, newend))
        if self.mode == 'local':
            self.results = self.query.slice(newstart, newend).all()
            if self.results:
                self._create_result_set(index=newstart)
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['getsubset']['url']
            params = {'searchfilter': self.search_filter, 'returnparams': self.return_params,
                      'start': newstart, 'end': newend, 'limit': chunk,
                      'sort': self.sortcol, 'order': self.order}
            self._interaction(url, params, calltype='getNext', create_set=True,
                              index=newstart)

        self.start = newstart
        self.end = newend
        self.count = len(self.results)

        if self.return_type:
            self.convertToTool(self.return_type)

        return self.results

    def getPrevious(self, chunk=None):
        ''' Retrieve the previous chunk of results.

        Returns a previous chunk of results from the query.
        from start to end in units of chunk.  Used with getNext
        to paginate through a long list of results

        Parameters:
            chunk (int):
                The number of objects to return

        Returns:
            results (list):
                A list of query results

        Example:
            >>> r = q.run()
            >>> r.getPrevious(5)
            >>> Retrieving previous 5, from 30 to 35
            >>> [(u'4-3988', u'1901', -9999.0),
            >>>  (u'4-3862', u'1902', -9999.0),
            >>>  (u'4-3293', u'1901', -9999.0),
            >>>  (u'4-3602', u'1902', -9999.0),
            >>>  (u'4-4602', u'1901', -9999.0)]

        See Also:
            getNext, getAll, getSubset

         '''

        if chunk and chunk < 0:
            warnings.warn('Chunk cannot be negative. Setting to {0}'.format(self.chunk), MarvinUserWarning)
            chunk = self.chunk

        newend = self.start
        self.chunk = chunk if chunk else self.chunk
        newstart = newend - self.chunk

        # This handles cases when the number of results is < total
        if self.totalcount == self.count:
            warnings.warn('You have all the results.  Cannot go back', MarvinUserWarning)
            return self.results

        # This handles the start edge case
        if newstart < 0:
            warnings.warn('You have reached the beginning.', MarvinUserWarning)
            newstart = 0
            newend = self.start

        # This grabs the previous chunk
        log.info('Retrieving previous {0}, from {1} to {2}'.format(self.chunk, newstart, newend))
        if self.mode == 'local':
            self.results = self.query.slice(newstart, newend).all()
            if self.results:
                self._create_result_set(index=newstart)
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['getsubset']['url']

            params = {'searchfilter': self.search_filter, 'returnparams': self.return_params,
                      'start': newstart, 'end': newend, 'limit': chunk,
                      'sort': self.sortcol, 'order': self.order}
            self._interaction(url, params, calltype='getPrevious', create_set=True,
                              index=newstart)

        self.start = newstart
        self.end = newend
        self.count = len(self.results)

        if self.return_type:
            self.convertToTool(self.return_type)

        return self.results

    def getSubset(self, start, limit=None):
        ''' Extracts a subset of results

        Parameters:
            start (int):
                The starting index of your subset extraction
            limit (int):
                The limiting number of results to return.

        Returns:
            results (list):
                A list of query results

        Example:
            >>> r = q.run()
            >>> r.getSubset(0, 10)
            >>> [(u'14-12', u'1901', -9999.0),
            >>> (u'14-13', u'1902', -9999.0),
            >>> (u'27-134', u'1901', -9999.0),
            >>> (u'27-100', u'1902', -9999.0),
            >>> (u'27-762', u'1901', -9999.0),
            >>> (u'27-759', u'1902', -9999.0),
            >>> (u'27-827', u'1901', -9999.0),
            >>> (u'27-828', u'1902', -9999.0),
            >>> (u'27-1170', u'1901', -9999.0),
            >>> (u'27-1167', u'1902', -9999.0)]

        See Also:
            getNext, getPrevious, getAll

        '''

        if not limit:
            limit = self.chunk

        if limit < 0:
            warnings.warn('Limit cannot be negative. Setting to {0}'.format(self.chunk), MarvinUserWarning)
            limit = self.chunk

        start = 0 if int(start) < 0 else int(start)
        end = start + int(limit)
        self.start = start
        self.end = end
        self.chunk = limit
        if self.mode == 'local':
            self.results = self.query.slice(start, end).all()
            if self.results:
                self._create_result_set(index=start)
        elif self.mode == 'remote':
            # Fail if no route map initialized
            if not config.urlmap:
                raise MarvinError('No URL Map found.  Cannot make remote call')

            # Get the query route
            url = config.urlmap['api']['getsubset']['url']

            params = {'searchfilter': self.search_filter, 'returnparams': self.return_params,
                      'start': start, 'end': end, 'limit': limit,
                      'sort': self.sortcol, 'order': self.order}
            self._interaction(url, params, calltype='getSubset', create_set=True, index=start)

        self.count = len(self.results)
        if self.return_type:
            self.convertToTool(self.return_type)

        return self.results

    def getAll(self, force=False):
        ''' Retrieve all of the results of a query

        Attempts to return all the results of a query.  The efficiency of this
        method depends heavily on how many rows and columns you wish to return.

        A cutoff limit is applied for results with more than 500,000 rows or
        results with more than 25 columns.

        Parameters
            force (bool):
                If True, force attempt to download everything

        Returns:
            The full list of query results.

        See Also:
            getNext, getPrevious, getSubset, loop

        '''

        if (self.totalcount > 500000 or len(self.columns) > 25) and not force:
            raise MarvinUserWarning("Cannot retrieve all results. The total number of requested "
                                    "rows or columns is too high. Please use the getNext, getPrevious, "
                                    "getSubset, or loop methods to retrieve pages.")

        if self.mode == 'local':
            self.results = self.query.from_self().all()
            self._create_result_set()
        elif self.mode == 'remote':
            # Get the query route
            url = config.urlmap['api']['querycubes']['url']

            params = {'searchfilter': self.search_filter, 'return_all': True,
                      'returnparams': self.return_params, 'limit': self.limit,
                      'sort': self.sortcol, 'order': self.order}
            self._interaction(url, params, calltype='getAll', create_set=True)
            self.count = self.totalcount
            print('Returned all {0} results'.format(self.totalcount))

    def convertToTool(self, tooltype, mode='auto', limit=None):
        ''' Converts the list of results into Marvin Tool objects

        Creates a list of Marvin Tool objects from a set of query results.
        The new list is stored in the Results.objects property.
        If the Query.returntype parameter is specified, then the Results object
        will automatically convert the results to the desired Tool on initialization.

        Parameters:
            tooltype (str):
                The requested Marvin Tool object that the results are converted into.
                Overrides the returntype parameter.  If not set, defaults
                to the returntype parameter.
            limit (int):
                Limit the number of results you convert to Marvin tools.  Useful
                for extremely large result sets.  Default is None.
            mode (str):
                The mode to use when attempting to convert to Tool. Default mode
                is to use the mode internal to Results. (most often remote mode)

        Example:
            >>> # Get the results from some query
            >>> r = q.run()
            >>> r.results
            >>> [NamedTuple(mangaid=u'14-12', name=u'1901', nsa.z=-9999.0),
            >>>  NamedTuple(mangaid=u'14-13', name=u'1902', nsa.z=-9999.0),
            >>>  NamedTuple(mangaid=u'27-134', name=u'1901', nsa.z=-9999.0),
            >>>  NamedTuple(mangaid=u'27-100', name=u'1902', nsa.z=-9999.0),
            >>>  NamedTuple(mangaid=u'27-762', name=u'1901', nsa.z=-9999.0)]

            >>> # convert results to Marvin Cube tools
            >>> r.convertToTool('cube')
            >>> r.objects
            >>> [<Marvin Cube (plateifu='7444-1901', mode='remote', data_origin='api')>,
            >>>  <Marvin Cube (plateifu='7444-1902', mode='remote', data_origin='api')>,
            >>>  <Marvin Cube (plateifu='7995-1901', mode='remote', data_origin='api')>,
            >>>  <Marvin Cube (plateifu='7995-1902', mode='remote', data_origin='api')>,
            >>>  <Marvin Cube (plateifu='8000-1901', mode='remote', data_origin='api')>]

        '''

        # set the desired tool type
        toollist = ['cube', 'spaxel', 'maps', 'rss', 'modelcube']
        tooltype = tooltype if tooltype else self.return_type
        assert tooltype in toollist, 'Returned tool type must be one of {0}'.format(toollist)

        # get the parameter list to check against
        paramlist = self.columns.full

        print('Converting results to Marvin {0} objects'.format(tooltype.title()))
        if tooltype == 'cube':
            self.objects = [self._get_object(Cube, plateifu=res.plateifu, mode=mode) for res in self.results[0:limit]]
        elif tooltype == 'maps':

            isbin = 'bintype.name' in paramlist
            istemp = 'template.name' in paramlist
            self.objects = []

            for res in self.results[0:limit]:
                mapkwargs = {'mode': mode, 'plateifu': res.plateifu}

                if isbin:
                    binval = res.bintype_name
                    mapkwargs['bintype'] = binval

                if istemp:
                    tempval = res.template_name
                    mapkwargs['template_kin'] = tempval

                self.objects.append(self._get_object(Maps, **mapkwargs))
        elif tooltype == 'spaxel':

            assert 'spaxelprop.x' in paramlist and 'spaxelprop.y' in paramlist, \
                   'Parameters must include spaxelprop.x and y in order to convert to Marvin Spaxel.'

            self.objects = []

            tab = self.toTable()
            uniq_plateifus = list(set(self.getListOf('plateifu')))

            for plateifu in uniq_plateifus:
                c = self._get_object(Cube, plateifu=plateifu, mode=mode)
                univals = tab['cube.plateifu'] == plateifu
                x = tab[univals]['spaxelprop.x'].tolist()
                y = tab[univals]['spaxelprop.y'].tolist()
                spaxels = c[y, x]
                self.objects.extend(spaxels)
        elif tooltype == 'rss':
            self.objects = [self._get_object(RSS, plateifu=res.plateifu, mode=mode) for res in self.results[0:limit]]
        elif tooltype == 'modelcube':

            isbin = 'bintype.name' in paramlist
            istemp = 'template.name' in paramlist
            self.objects = []

            assert self.release != 'MPL-4', "ModelCubes require a release of MPL-5 and up"

            for res in self.results[0:limit]:
                mapkwargs = {'mode': mode, 'plateifu': res.plateifu}

                if isbin:
                    binval = res.bintype_name
                    mapkwargs['bintype'] = binval

                if istemp:
                    tempval = res.template_name
                    mapkwargs['template_kin'] = tempval

                self.objects.append(self._get_object(ModelCube, **mapkwargs))

    @staticmethod
    def _get_object(obj, **kwargs):
        ''' Return a Marvin object or an error message

        To preserve the lengths of self.results and self.objects, it will
        either return an instance or an error message in its place

        Parameters:
            obj (object):
                A Marvin Class object
            kwargs:
                Any set of parameters to instantiate a Marvin object

        Returns:
            The Marvin instance or an error message if it failed
        '''

        try:
            inst = obj(**kwargs)
        except MarvinError as e:
            plateifu = kwargs.get('plateifu', '')
            inst = 'Error creating {0} for {1}: {2}'.format(obj.__name__, plateifu, e)

        return inst

    def plot(self, x_name, y_name, **kwargs):
        ''' Make a scatter plot from two columns of results

        Creates a Matplotlib scatter plot from Results columns.
        Accepts as input two string column names.  Will extract the total
        entire column (if not already available) and plot them.  Creates
        a scatter plot with (optionally) adjoining 1-d histograms for each column.

        See :meth:`marvin.utils.plot.scatter.plot` and
        :meth:`marvin.utils.plot.scatter.hist` for details.

        Parameters:
            x_name (str):
                The name of the x-column of data. Required
            y_name (str):
                The name of the y-column of data. Required
            return_plateifus (bool):
                If True, includes the plateifus in each histogram bin in the
                histogram output.  Default is True.
            return_figure (bool):
                Set to False to not return the Figure and Axis object. Defaults to True.
            show_plot (bool):
                Set to False to not show the interactive plot
            **kwargs (dict):
                Any other keyword argument that will be passed to Marvin's
                scatter and hist plotting methods

        Returns:
            The figure, axes, and histogram data from the plotting function

        Example:
            >>> # do a query and get the results
            >>> q = Query(search_filter='nsa.z < 0.1', returnparams=['nsa.elpetro_ba', 'g_r'])
            >>> r = q.run()
            >>> # plot the total columns of Redshift vs g-r magnitude
            >>> fig, axes, hist_data = r.plot('nsa.z', 'g_r')

        '''

        assert all([x_name, y_name]), 'Must provide both an x and y column'
        return_plateifus = kwargs.pop('return_plateifus', True)
        with_hist = kwargs.get('with_hist', True)
        show_plot = kwargs.pop('show_plot', True)
        return_figure = kwargs.get('return_figure', True)

        # get the named column
        x_col = self.columns[x_name]
        y_col = self.columns[y_name]

        # get the values of the two columns
        if self.count != self.totalcount:
            x_data = self.getListOf(x_name, return_all=True)
            y_data = self.getListOf(y_name, return_all=True)
        else:
            x_data = self.results[x_name]
            y_data = self.results[y_name]

        with turn_off_ion(show_plot=show_plot):
            output = marvin.utils.plot.scatter.plot(x_data, y_data, xlabel=x_col, ylabel=y_col, **kwargs)

        # computes a list of plateifus in each bin
        if return_plateifus and with_hist:
            plateifus = self.getListOf('plateifu', return_all=True)
            hdata = output[2] if return_figure else output
            if 'xhist' in hdata:
                hdata['xhist']['bins_plateifu'] = map_bins_to_column(plateifus, hdata['xhist']['indices'])
            if 'yhist' in hdata:
                hdata['yhist']['bins_plateifu'] = map_bins_to_column(plateifus, hdata['yhist']['indices'])
            output = output[0:2] + (hdata,) if return_figure else hdata

        return output

    def hist(self, name, **kwargs):
        ''' Make a histogram for a given column of the results

        Creates a Matplotlib histogram from a Results Column.
        Accepts as input a string column name.  Will extract the total
        entire column (if not already available) and plot it.

        See :meth:`marvin.utils.plot.scatter.hist` for details.

        Parameters:
            name (str):
                The name of the column of data. Required
            return_plateifus (bool):
                If True, includes the plateifus in each histogram bin in the
                histogram output.  Default is True.
            return_figure (bool):
                Set to False to not return the Figure and Axis object. Defaults to True.
            show_plot (bool):
                Set to False to not show the interactive plot
            **kwargs (dict):
                Any other keyword argument that will be passed to Marvin's
                hist plotting methods

        Returns:
            The histogram data, figure, and axes from the plotting function

        Example:
            >>> # do a query and get the results
            >>> q = Query(search_filter='nsa.z < 0.1', returnparams=['nsa.elpetro_ba', 'g_r'])
            >>> r = q.run()
            >>> # plot a histogram of the redshift column
            >>> hist_data, fig, axes = r.hist('nsa.z')

        '''

        return_plateifus = kwargs.pop('return_plateifus', True)
        show_plot = kwargs.pop('show_plot', True)
        return_figure = kwargs.get('return_figure', True)

        # get the named column
        col = self.columns[name]

        # get the values of the two columns
        if self.count != self.totalcount:
            data = self.getListOf(name, return_all=True)
        else:
            data = self.results[name]

        # xhist, fig, ax_hist_x = output
        with turn_off_ion(show_plot=show_plot):
            output = marvin.utils.plot.scatter.hist(data, **kwargs)

        if return_plateifus:
            plateifus = self.getListOf('plateifu', return_all=True)
            hdata = output[0] if return_figure else output
            hdata['bins_plateifu'] = map_bins_to_column(plateifus, hdata['indices'])
            output = (hdata,) + output[1:] if return_figure else hdata

        return output



#!/usr/bin/python

# -------------------------------------------------------------------
# Import statements
# -------------------------------------------------------------------
import math
import os
import re
import sys
from decimal import *
from operator import *

import marvin.db.models.SampleModelClasses as sampledb
import numpy as np
from astropy.io import fits
from flask_login import UserMixin
from marvin.core.caching_query import RelationshipCache
from marvin.db.ArrayUtils import ARRAY_D
from marvin.db.database import db
from sqlalchemy import and_, func, select  # for aggregate, other functions
from sqlalchemy.dialects.postgresql import *
from sqlalchemy.engine import reflection
from sqlalchemy.ext.hybrid import hybrid_method, hybrid_property

from sqlalchemy.orm import configure_mappers, deferred, relationship
from sqlalchemy.orm.session import Session
from sqlalchemy.schema import Column
from sqlalchemy.sql import column
from sqlalchemy.types import JSON, Float, Integer, String
from sqlalchemy_utils import Timestamp
from werkzeug.security import check_password_hash, generate_password_hash

try:
    from sdss_access.path import Path
except ImportError as e:
    Path = None



# ========================
# Define database classes
# ========================
Base = db.Base


class ArrayOps(object):
    ''' this class adds array functionality '''

    __tablename__ = 'arrayops'
    __table_args__ = {'extend_existing': True}

    @property
    def cols(self):
        return list(self.__table__.columns._data.keys())

    @property
    def collist(self):
        return ['wavelength', 'flux', 'ivar', 'mask', 'xpos', 'ypos', 'specres']

    def getTableName(self):
        return self.__table__.name

    def matchIndex(self, name=None):

        # Get index of correct column
        incols = [x for x in self.cols if x in self.collist]
        if not any(incols):
            return None
        elif len(incols) == 1:
            idx = self.cols.index(incols[0])
        else:
            if not name:
                print('Multiple columns found.  Column name must be specified!')
                return None
            elif name in self.collist:
                idx = self.cols.index(name)
            else:
                return None

        return idx

    def filter(self, start, end, name=None):

        # Check input types or map string operators
        startnum = type(start) == int or type(start) == float
        endnum = type(end) == int or type(end) == float
        opdict = {'=': eq, '<': lt, '<=': le, '>': gt, '>=': ge, '!=': ne}
        if start in opdict.keys() or end in opdict.keys():
            opind = list(opdict.keys()).index(start) if start in opdict.keys() else list(opdict.keys()).index(end)
            if start in opdict.keys():
                start = opdict[list(opdict.keys())[opind]]
            if end in opdict.keys():
                end = opdict[list(opdict.keys())[opind]]

        # Get matching index
        self.idx = self.matchIndex(name=name)
        if not self.idx:
            return None

        # Perform calculation
        try:
            data = self.__getattribute__(self.cols[self.idx])
        except:
            data = None

        if data:
            if startnum and endnum:
                arr = [x for x in data if x >= start and x <= end]
            elif not startnum and endnum:
                arr = [x for x in data if start(x, end)]
            elif startnum and not endnum:
                arr = [x for x in data if end(x, start)]
            elif startnum == eq or endnum == eq:
                arr = [x for x in data if start(x, end)] if start == eq else [x for x in data if end(x, start)]
            return arr
        else:
            return None

    def equal(self, num, name=None):

        # Get matching index
        self.idx = self.matchIndex(name=name)
        if not self.idx:
            return None

        # Perform calculation
        try:
            data = self.__getattribute__(self.cols[self.idx])
        except:
            data = None

        if data:
            arr = [x for x in data if x == num]
            return arr
        else:
            return None

    def less(self, num, name=None):

        # Get matching index
        self.idx = self.matchIndex(name=name)
        if not self.idx:
            return None

        # Perform calculation
        try:
            data = self.__getattribute__(self.cols[self.idx])
        except:
            data = None

        if data:
            arr = [x for x in data if x <= num]
            return arr
        else:
            return None

    def greater(self, num, name=None):

        # Get matching index
        self.idx = self.matchIndex(name=name)
        if not self.idx:
            return None

        # Perform calculation
        try:
            data = self.__getattribute__(self.cols[self.idx])
        except:
            data = None

        if data:
            arr = [x for x in data if x >= num]
            return arr
        else:
            return None

    def getIndices(self, arr):

        if self.idx:
            indices = [self.__getattribute__(self.cols[self.idx]).index(a) for a in arr]
        else:
            return None

        return indices


class Cube(Base, ArrayOps):
    __tablename__ = 'cube'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb', 'extend_existing': True}

    specres = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    specresd = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    prespecres = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    prespecresd = deferred(Column(ARRAY_D(Float, zero_indexes=True)))

    def __repr__(self):
        return '<Cube (pk={0}, plate={1}, ifudesign={2}, tag={3})>'.format(self.pk, self.plate, self.ifu.name, self.pipelineInfo.version.version)

    @property
    def header(self):
        '''Returns an astropy header'''

        session = Session.object_session(self)
        data = session.query(FitsHeaderKeyword.label, FitsHeaderValue.value,
                             FitsHeaderValue.comment).join(FitsHeaderValue).filter(
            FitsHeaderValue.cube == self).all()

        hdr = fits.Header(data)
        return hdr

    @property
    def name(self):
        return 'manga-{0}-{1}-LOGCUBE.fits.gz'.format(self.plate, self.ifu.name)

    @property
    def default_mapsname(self):
        return 'mangadap-{0}-{1}-default.fits.gz'.format(self.plate, self.ifu.name)

    def getPath(self):
        sasurl = os.getenv('SAS_URL')

        if sasurl:
            sasredux = os.path.join(sasurl, 'sas/mangawork/manga/spectro/redux')
            path = sasredux
        else:
            redux = os.getenv('MANGA_SPECTRO_REDUX')
            path = redux

        version = self.pipelineInfo.version.version
        cubepath = os.path.join(path, version, str(self.plate), 'stack')
        return cubepath

    @property
    def location(self):
        name = self.name
        path = self.getPath()
        loc = os.path.join(path, name)
        return loc

    @property
    def image(self):
        ifu = '{0}.png'.format(self.ifu.name)
        path = self.getPath()
        imageloc = os.path.join(path, 'images', ifu)
        return imageloc

    def header_to_dict(self):
        '''Returns a simple python dictionary header'''

        values = self.headervals
        hdrdict = {str(val.keyword.label): val.value for val in values}
        return hdrdict

    @property
    def plateclass(self):
        '''Returns a plate class'''

        plate = Plate(self)

        return plate

    def testhead(self, key):
        ''' Test existence of header keyword'''

        try:
            if self.header_to_dict()[key]:
                return True
        except:
            return False

    def getFlags(self, bits, name):
        session = Session.object_session(self)

        # if bits not a digit, return None
        if not str(bits).isdigit():
            return 'NULL'
        else:
            bits = int(bits)

        # Convert the integer value to list of bits
        bitlist = [int(i) for i in '{0:08b}'.format(bits)]
        bitlist.reverse()
        indices = [i for i, bit in enumerate(bitlist) if bit]

        labels = []
        for i in indices:
            maskbit = session.query(MaskBit).filter_by(flag=name, bit=i).one()
            labels.append(maskbit.label)

        return labels

    def getQualBits(self, stage='3d'):
        ''' get quality flags '''

        col = 'DRP2QUAL' if stage == '2d' else 'DRP3QUAL'
        hdr = self.header_to_dict()
        bits = hdr.get(col, None)
        return bits

    def getQualFlags(self, stage='3d'):
        ''' get quality flags '''

        name = 'MANGA_DRP2QUAL' if stage == '2d' else 'MANGA_DRP3QUAL'
        bits = self.getQualBits(stage=stage)

        if bits:
            return self.getFlags(bits, name)
        else:
            return None

    def getTargFlags(self, targtype=1):
        ''' get target flags '''

        name = 'MANGA_TARGET1' if targtype == 1 else 'MANGA_TARGET2' if targtype == 2 else 'MANGA_TARGET3'
        bits = self.getTargBits(targtype=targtype)
        if bits:
            return self.getFlags(bits, name)
        else:
            return None

    def getTargBits(self, targtype=1):
        ''' get target bits '''

        assert targtype in [1,2,3], 'target type can only 1, 2 or 3'

        hdr = self.header_to_dict()
        newcol = 'MNGTARG{0}'.format(targtype)
        oldcol = 'MNGTRG{0}'.format(targtype)
        bits = hdr.get(newcol, hdr.get(oldcol, None))
        return bits

    def get3DCube(self, extension='flux'):
        """Returns a 3D array of ``extension`` from the cube spaxels.

        For example, ``cube.get3DCube('flux')`` will return the original
        flux cube with the same ordering as the FITS data cube.

        Note that this method seems to be really slow retrieving arrays (this
        is especially serious for large IFUs).

        """

        session = Session.object_session(self)
        spaxels = session.query(getattr(Spaxel, extension)).filter(
            Spaxel.cube_pk == self.pk).order_by(Spaxel.x, Spaxel.y).all()

        # Assumes cubes are always square (!)
        nx = ny = int(np.sqrt(len(spaxels)))
        nwave = len(spaxels[0][0])

        spArray = np.array(spaxels)

        return spArray.transpose().reshape((nwave, ny, nx)).transpose(0, 2, 1)

    @hybrid_property
    def plateifu(self):
        '''Returns parameter plate-ifu'''
        return '{0}-{1}'.format(self.plate, self.ifu.name)

    @plateifu.expression
    def plateifu(cls):
        return func.concat(Cube.plate, '-', IFUDesign.name)

    @hybrid_property
    def restwave(self):
        if self.target:
            redshift = self.target.NSA_objects[0].z
            wave = np.array(self.wavelength.wavelength)
            restwave = wave / (1 + redshift)
            return restwave
        else:
            return None

    @restwave.expression
    def restwave(cls):
        restw = (func.rest_wavelength(sampledb.NSA.z))
        return restw

    def has_modelspaxels(self, name=None):
        if not name:
            name = '(SPX|HYB)'
        has_ms = False
        model_cubes = [f.modelcube for f in self.dapfiles if re.search('LOGCUBE-{0}'.format(name), f.filename)]
        if model_cubes:
            mc = sum(model_cubes, [])
            if mc:
                from marvin.db.models.DapModelClasses import ModelSpaxel
                session = Session.object_session(mc[0])
                ms = session.query(ModelSpaxel).filter_by(modelcube_pk=mc[0].pk).first()
                has_ms = True if ms else False
        return has_ms

    def has_spaxels(self):
        if len(self.spaxels) > 0:
            return True
        else:
            return False

    def has_fibers(self):
        if len(self.fibers) > 0:
            return True
        else:
            return False


def set_quality(stage):
    ''' produces cube quality flag '''

    col = 'DRP2QUAL' if stage == '2d' else 'DRP3QUAL'
    label = 'cubequal{0}'.format(stage)
    kwarg = 'DRP{0}QUAL'.format(stage[0])

    @hybrid_property
    def quality(self):
        bits = self.getQualBits(stage=stage)
        return int(bits)

    @quality.expression
    def quality(cls):
        return select([FitsHeaderValue.value.cast(Integer)]).\
                      where(and_(FitsHeaderKeyword.pk==FitsHeaderValue.fits_header_keyword_pk,
                                 FitsHeaderKeyword.label.ilike(kwarg),
                                 FitsHeaderValue.cube_pk==cls.pk)).\
                      label(label)
    return quality


def set_manga_target(targtype):
    ''' produces manga_target flags '''

    label = 'mngtrg{0}'.format(targtype)
    kwarg = 'MNGT%RG{0}'.format(targtype)

    @hybrid_property
    def target(self):
        bits = self.getTargBits(targtype=targtype)
        return int(bits)

    @target.expression
    def target(cls):
        return select([FitsHeaderValue.value.cast(Integer)]).\
                      where(and_(FitsHeaderKeyword.pk==FitsHeaderValue.fits_header_keyword_pk,
                                 FitsHeaderKeyword.label.ilike(kwarg),
                                 FitsHeaderValue.cube_pk==cls.pk)).\
                      label(label)
    return target

setattr(Cube, 'manga_target1', set_manga_target(1))
setattr(Cube, 'manga_target2', set_manga_target(2))
setattr(Cube, 'manga_target3', set_manga_target(3))
setattr(Cube, 'quality', set_quality('3d'))


class Wavelength(Base, ArrayOps):
    __tablename__ = 'wavelength'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb', 'extend_existing': True}

    wavelength = deferred(Column(ARRAY_D(Float, zero_indexes=True)))

    def __repr__(self):
        return '<Wavelength (pk={0})>'.format(self.pk)


class Spaxel(Base, ArrayOps):
    __tablename__ = 'spaxel'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb', 'extend_existing': True}

    flux = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    ivar = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    mask = deferred(Column(ARRAY_D(Integer, zero_indexes=True)))
    disp = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    predisp = deferred(Column(ARRAY_D(Float, zero_indexes=True)))

    def __repr__(self):
        return '<Spaxel (pk={0}, x={1}, y={2})'.format(self.pk, self.x, self.y)

    @hybrid_method
    def sum(self, name=None):

        total = sum(self.flux)

        return total

    @sum.expression
    def sum(cls):
        # return select(func.sum(func.unnest(cls.flux))).select_from(func.unnest(cls.flux)).label('totalflux')
        return select([func.sum(column('totalflux'))]).select_from(func.unnest(cls.flux).alias('totalflux'))


class RssFiber(Base, ArrayOps):
    __tablename__ = 'rssfiber'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb', 'extend_existing': True}

    flux = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    ivar = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    mask = deferred(Column(ARRAY_D(Integer, zero_indexes=True)))
    xpos = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    ypos = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    disp = deferred(Column(ARRAY_D(Float, zero_indexes=True)))
    predisp = deferred(Column(ARRAY_D(Float, zero_indexes=True)))

    def __repr__(self):
        return '<RssFiber (pk={0}, expnum={1}, mjd={2}, fiber={3})>'.format(self.pk, self.exposure_no, self.mjd, self.fiber.fiberid)


class PipelineInfo(Base):
    __tablename__ = 'pipeline_info'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return ('<Pipeline_Info (pk={0}, name={3}, ver={1}, release={2})>'.format(self.pk, 
                self.version.version, self.version.label, self.name.label))


class PipelineVersion(Base):
    __tablename__ = 'pipeline_version'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Pipeline_Version (pk={0}, version={1}, release={2})>'.format(self.pk, self.version, self.label)


class PipelineStage(Base):
    __tablename__ = 'pipeline_stage'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Pipeline_Stage (pk={0}, label={1})>'.format(self.pk, self.label)


class PipelineCompletionStatus(Base):
    __tablename__ = 'pipeline_completion_status'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Pipeline_Completion_Status (pk={0}, label={1})>'.format(self.pk, self.label)


class PipelineName(Base):
    __tablename__ = 'pipeline_name'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Pipeline_Name (pk={0}, label={1})>'.format(self.pk, self.label)


class FitsHeaderValue(Base):
    __tablename__ = 'fits_header_value'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Fits_Header_Value (pk={0})'.format(self.pk)


class FitsHeaderKeyword(Base):
    __tablename__ = 'fits_header_keyword'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Fits_Header_Keyword (pk={0}, label={1})'.format(self.pk, self.label)


class IFUDesign(Base):
    __tablename__ = 'ifudesign'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<IFU_Design (pk={0}, name={1})'.format(self.pk, self.name)

    @property
    def ifuname(self):
        return self.name

    @property
    def designid(self):
        return self.name

    @property
    def ifutype(self):
        return self.name[:-2]


class IFUToBlock(Base):
    __tablename__ = 'ifu_to_block'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<IFU_to_Block (pk={0})'.format(self.pk)


class SlitBlock(Base):
    __tablename__ = 'slitblock'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<SlitBlock (pk={0})'.format(self.pk)


class Cart(Base):
    __tablename__ = 'cart'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Cart (pk={0}, id={1})'.format(self.pk, self.id)


class Fibers(Base):
    __tablename__ = 'fibers'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Fibers (pk={0}, fiberid={1}, fnum={2})'.format(self.pk, self.fiberid, self.fnum)


class FiberType(Base):
    __tablename__ = 'fiber_type'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Fiber_Type (pk={0},label={1})'.format(self.pk, self.label)


class TargetType(Base):
    __tablename__ = 'target_type'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Target_Type (pk={0},label={1})'.format(self.pk, self.label)


class Sample(Base, ArrayOps):
    __tablename__ = 'sample'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<Sample (pk={0},cube={1})'.format(self.pk, self.cube)

    @hybrid_property
    def nsa_logmstar(self):
        try:
            return math.log10(self.nsa_mstar)
        except ValueError:
            return -9999.0
        except TypeError:
            return None

    @nsa_logmstar.expression
    def nsa_logmstar(cls):
        return func.log(cls.nsa_mstar)

    @hybrid_property
    def nsa_logmstar_el(self):
        try:
            return math.log10(self.nsa_mstar_el)
        except ValueError as e:
            return -9999.0
        except TypeError as e:
            return None

    @nsa_logmstar_el.expression
    def nsa_logmstar_el(cls):
        return func.log(cls.nsa_mstar_el)


class CartToCube(Base):
    __tablename__ = 'cart_to_cube'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<CartToCube (pk={0},cube={1}, cart={2})'.format(self.pk, self.cube, self.cart)


class Wcs(Base, ArrayOps):
    __tablename__ = 'wcs'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<WCS (pk={0},cube={1})'.format(self.pk, self.cube)

    def makeHeader(self):
        wcscols = self.cols[2:]
        newhdr = fits.Header()
        for c in wcscols:
            newhdr[c] = float(self.__getattribute__(c)) if type(self.__getattribute__(c)) == Decimal else self.__getattribute__(c)
        return newhdr


class ObsInfo(Base):
    __tablename__ = 'obsinfo'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    _expnum = Column('expnum', String)

    @hybrid_property
    def expnum(self):
        return func.trim(self._expnum)

    def __repr__(self):
        return '<ObsInfo (pk={0},cube={1})'.format(self.pk, self.cube)


class CubeShape(Base):
    __tablename__ = 'cube_shape'
    __table_args__ = {'autoload': True, 'schema': 'mangadatadb'}

    def __repr__(self):
        return '<CubeShape (pk={0},cubes={1},size={2},totalrows={3})'.format(self.pk, len(self.cubes), self.size, self.total)

    @property
    def shape(self):
        return (self.size, self.size)

    def makeIndiceArray(self):
        ''' Return the indices array as a numpy array '''
        return np.array(self.indices)

    def getXY(self, index=None):
        ''' Get the x,y elements from a single digit index '''
        if index is not None:
            if index > self.total:
                return None, None
            else:
                i = int(index / self.size)
                j = int(index - i * self.size)
        else:
            arrind = self.makeIndiceArray()
            i = np.array(arrind / self.size, dtype=int)
            j = np.array(self.makeIndiceArray() - i * self.size, dtype=int)
        return i, j

    @hybrid_property
    def x(self):
        '''Returns parameter plate-ifu'''
        x = self.getXY()[0]
        return x

    @x.expression
    def x(cls):
        #arrind = func.unnest(cls.indices).label('arrind')
        #x = func.array_agg(arrind / cls.size).label('x')
        s = db.Session()
        arrind = (func.unnest(cls.indices) / cls.size).label('xarrind')
        #x = s.query(arrind).select_from(cls).subquery('xarr')
        #xagg = s.query(func.array_agg(x.c.xarrind))
        return arrind

    @hybrid_property
    def y(self):
        '''Returns parameter plate-ifu'''
        y = self.getXY()[1]
        return y

    @y.expression
    def y(cls):
        #arrind = func.unnest(cls.indices).label('arrind')
        #x = arrind / cls.size
        #y = func.array_agg(arrind - x*cls.size).label('y')
        #return y
        s = db.Session()
        arrunnest = func.unnest(cls.indices)
        xarr = (func.unnest(cls.indices) / cls.size).label('xarrind')
        arrind = (arrunnest - xarr*cls.size).label('yarrind')
        #n.arrind-(n.arrind/n.size)*n.size
        y = s.query(arrind).select_from(cls).subquery('yarr')
        yagg = s.query(func.array_agg(y.c.yarrind))
        return yagg.as_scalar()


class Plate(object):
    ''' new plate class '''

    __tablename__ = 'myplate'

    def __init__(self, cube=None, id=None):
        self.id = cube.plate if cube else id if id else None
        self.cube = cube if cube else None
        self.drpver = None
        if self.cube:
            self._hdr = self.cube.header_to_dict()
            self.type = self.getPlateType()
            self.platetype = self._hdr.get('PLATETYP', None)
            self.surveymode = self._hdr.get('SRVYMODE', None)
            self.dateobs = self._hdr.get('DATE-OBS', None)
            self.ra = self._hdr.get('CENRA', None)
            self.dec = self._hdr.get('CENDEC', None)
            self.designid = self._hdr.get('DESIGNID', None)
            self.cartid = self._hdr.get('CARTID', None)
            self.drpver = self.cube.pipelineInfo.version.version
            self.isbright = 'APOGEE' in self.surveymode
            self.dir3d = 'mastar' if self.isbright else 'stack'

            # cast a few
            self.ra = float(self.ra) if self.ra else None
            self.dec = float(self.dec) if self.dec else None
            self.id = int(self.id) if self.id else None
            self.designid = int(self.designid) if self.designid else None

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return ('Plate (id={0}, ra={1}, dec={2}, '
                ' designid={3})'.format(self.id, self.ra, self.dec, self.designid))

    def getPlateType(self):
        ''' Get the type of MaNGA plate '''

        hdr = self.cube.header

        # try galaxy
        mngtrg = self._hdr.get('MNGTRG1', None)
        pltype = 'Galaxy' if mngtrg else None

        # try stellar
        if not pltype:
            mngtrg = self._hdr.get('MNGTRG2', None)
            pltype = 'Stellar' if mngtrg else None

        # try ancillary
        if not pltype:
            mngtrg = self._hdr.get('MNGTRG3', None)
            pltype = 'Ancillary' if mngtrg else None

        return pltype

    @property
    def cubes(self):
        ''' Get all cubes on this plate '''

        session = db.Session()
        if self.drpver:
            cubes = session.query(Cube).join(PipelineInfo, PipelineVersion).\
                filter(Cube.plate == self.id, PipelineVersion.version == self.drpver).all()
        else:
            cubes = session.query(Cube).filter(Cube.plate == self.id).all()
        return cubes


# ================
# manga Aux DB classes
# ================
class CubeHeader(Base):
    __tablename__ = 'cube_header'
    __table_args__ = {'autoload': True, 'schema': 'mangaauxdb'}

    header = Column(JSON)

    def __repr__(self):
        return '<CubeHeader (pk={0},cube={1})'.format(self.pk, self.cube)


class MaskLabels(Base):
    __tablename__ = 'maskbit_labels'
    __table_args__ = {'autoload': True, 'schema': 'mangaauxdb'}

    def __repr__(self):
        return '<MaskLabels (pk={0},bit={1})'.format(self.pk, self.maskbit)


class MaskBit(Base):
    __tablename__ = 'maskbit'
    __table_args__ = {'autoload': True, 'schema': 'mangaauxdb'}

    def __repr__(self):
        return '<MaskBit (pk={0},flag={1}, bit={2}, label={3})'.format(self.pk, self.flag, self.bit, self.label)


# ================
# Query Meta classes
# ================
class QueryMeta(Base, Timestamp):
    __tablename__ = 'query'
    __table_args__ = {'autoload': True, 'schema': 'history'}

    def __repr__(self):
        return '<QueryMeta (pk={0}, filter={1}), count={2}>'.format(self.pk, self.searchfilter, self.count)


class User(Base, UserMixin, Timestamp):
    __tablename__ = 'user'
    __table_args__ = {'autoload': True, 'schema': 'history'}

    def __repr__(self):
        return '<User (pk={0}, username={1})'.format(self.pk, self.username)

    def set_password(self, password):
        self.password_hash = generate_password_hash(password)

    def check_password(self, password):
        return check_password_hash(self.password_hash, password)

    def get_id(self):
        return (self.pk)

    def update_stats(self, request=None):
        remote_addr = request.remote_addr or None
        self.login_count += 1
        old_current_ip, new_current_ip = self.current_ip, remote_addr
        self.last_ip = old_current_ip
        self.current_ip = new_current_ip

# Define relationships
# ========================

Cube.pipelineInfo = relationship(PipelineInfo, backref="cubes")
Cube.wavelength = relationship(Wavelength, backref="cube")
Cube.ifu = relationship(IFUDesign, backref="cubes")
Cube.carts = relationship(Cart, secondary=CartToCube.__table__, backref="cubes")
Cube.wcs = relationship(Wcs, backref='cube', uselist=False)
Cube.shape = relationship(CubeShape, backref='cubes', uselist=False)
Cube.obsinfo = relationship(ObsInfo, backref='cube', uselist=False)

# from SampleDB
Cube.target = relationship(sampledb.MangaTarget, backref='cubes')

Sample.cube = relationship(Cube, backref="sample", uselist=False)

FitsHeaderValue.cube = relationship(Cube, backref="headervals")
FitsHeaderValue.keyword = relationship(FitsHeaderKeyword, backref="value")

IFUDesign.blocks = relationship(SlitBlock, secondary=IFUToBlock.__table__, backref='ifus')
Fibers.ifu = relationship(IFUDesign, backref="fibers")
Fibers.fibertype = relationship(FiberType, backref="fibers")
Fibers.targettype = relationship(TargetType, backref="fibers")

insp = reflection.Inspector.from_engine(db.engine)
fks = insp.get_foreign_keys(Spaxel.__table__.name, schema='mangadatadb')
if fks:
    Spaxel.cube = relationship(Cube, backref='spaxels')
fks = insp.get_foreign_keys(RssFiber.__table__.name, schema='mangadatadb')
if fks:
    RssFiber.cube = relationship(Cube, backref='rssfibers')
    RssFiber.fiber = relationship(Fibers, backref='rssfibers')

PipelineInfo.name = relationship(PipelineName, backref="pipeinfo")
PipelineInfo.stage = relationship(PipelineStage, backref="pipeinfo")
PipelineInfo.version = relationship(PipelineVersion, backref="pipeinfo")
PipelineInfo.completionStatus = relationship(PipelineCompletionStatus, backref="pipeinfo")

# from AuxDB
CubeHeader.cube = relationship(Cube, backref='hdr')

# ---------------------------------------------------------
# Test that all relationships/mappings are self-consistent.
# ---------------------------------------------------------

try:
    configure_mappers()
except RuntimeError as error:
    print("""
mangadb.DataModelClasses:
An error occurred when verifying the relationships between the database tables.
Most likely this is an error in the definition of the SQLAlchemy relationships -
see the error message below for details.
""")
    print("Error type: %s" % sys.exc_info()[0])
    print("Error value: %s" % sys.exc_info()[1])
    print("Error trace: %s" % sys.exc_info()[2])
    sys.exit(1)


data_cache = RelationshipCache(Cube.target).\
               and_(RelationshipCache(Cube.pipelineInfo)).\
               and_(RelationshipCache(Cube.ifu)).\
               and_(RelationshipCache(Cube.spaxels)).\
               and_(RelationshipCache(Cube.wavelength)).\
               and_(RelationshipCache(Cube.wcs)).\
               and_(RelationshipCache(Cube.shape)).\
               and_(RelationshipCache(Cube.obsinfo)).\
               and_(RelationshipCache(IFUDesign.fibers)).\
               and_(RelationshipCache(PipelineInfo.version)).\
               and_(RelationshipCache(Cube.rssfibers))



# !usr/bin/env python2
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-06-28 15:31:51
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2018-06-05 11:39:50

from __future__ import print_function, division, absolute_import
from flask_compress import Compress
from flask_featureflags import FeatureFlag
from raven.contrib.flask import Sentry
from flask_limiter import Limiter
from flask_limiter.util import get_remote_address
from flask_profiler import Profiler
from flask_caching import Cache
from flask_login import LoginManager
from flask_jwt_extended import JWTManager
from flask_cors import CORS
from flask_session import Session
import flask_jsglue as jsg
import logging

# JS Glue (allows use of Flask.url_for inside javascript)
jsglue = jsg.JSGlue()

# Feature Flags (allows turning on/off of features)
flags = FeatureFlag()

# Sentry error logging
sentry = Sentry(logging=True, level=logging.ERROR)

# Route Rate Limiter
limiter = Limiter(key_func=get_remote_address)

# Flask Profiler
profiler = Profiler()

# Flask Cache
cache = Cache()

# Flask Login
login_manager = LoginManager()

# Flask-JWT (JSON Web Token)
jwt = JWTManager()

# Flask CORS
cors = CORS()

# Flask Session
session = Session()

# Flask Compress
compress = Compress()



# !/usr/bin/env python
# -*- coding: utf-8 -*-
# 
# Filename: caching.py
# Project: mixins
# Author: Brian Cherinka
# Created: Monday, 27th January 2020 2:41:01 pm
# License: BSD 3-clause "New" or "Revised" License
# Copyright (c) 2020 Brian Cherinka
# Last Modified: Monday, 27th January 2020 6:35:50 pm
# Modified By: Brian Cherinka


from __future__ import print_function, division, absolute_import
from marvin import config
from marvin.db.caching import regions

__all__ = ['CacheMixIn']


cache_dict = {'Map': 'maps', 'Maps': 'maps', 'ModelCube': 'models'}


class CacheMixIn(object):
    ''' A mixin that provides optional caching of database tools
    
    When mixed in, determines the cache region to use based on the name
    of the class it's mixed with.  Can turn off caching altogether by setting
    use_db_tools_cache: False in custom marvin config yaml file.  If set,
    uses a null region cache.    
    '''
    _cache_regions = list(regions.keys())

    def __init__(self, region='default'):
        assert region in self._cache_regions, 'region must be a proper cache region'

        class_name = self.__class__.__name__

        use_tool_cache = config._custom_config.get('use_db_tools_cache', True)
        cache_region = 'null' if not use_tool_cache else cache_dict.get(
            class_name, region)
        self.cache_region = cache_region

